#!/usr/local/bin/perl
# -*-Perl-*-
# Last changed Time-stamp: <2004-01-20 14:17:35 xtina>

# reads bracket-format file of the alignment  from stdin and writes
# bracket-format of the sequence specified with seqname to stdout
# basepairs not in the allowed list (AU, UA, GC, CG, GU, UG) are removed

# usage: alndb2seqdb.pl -n <seqname> <aln_dotbracket-file> <aln-file>

# the right order of the files is mandatory
# the seqname has to be the same as in the alignment file
# <aln_dotbracket-file> is for example the output file of hxmatch (name_db.hx)
# <aln-file> has to be in CLUSTAL format

use Getopt::Std;
use strict;
use vars qw/$stru @file $name @number $opt_n @alnseq @file  $seq @allow/;

@allow = qw(AU UA GC CG GU UG);

getopt('n');
$name = $opt_n;

my $i=1;
while (my $fname=shift(@ARGV)) {  
  $file[$i++] = $fname ;
}

die("Need a sequence name!\n") if !$name;
#open files
open(DB, "< $file[1]") || die("Can't open '$file[1]'!!!\n");
open(ALN, "< $file[2]") || die("Can't open '$file[2]!!!\n");

#read alignment
while (my $line = <ALN>) {
  if ($line =~ m/$name/) {
     $line =~ s/$name//;
    chomp $line;
    $line =~ tr/a-z/A-Z/;    
    $line =~ s/\s+//;
    push @alnseq, split //, $line;
  }
}
my $len = $#alnseq;
foreach (@alnseq) {
  $seq .= $_ if $_ ne '-'; 
}
 my $seqlen = length $seq;
#print "$seqlen\n";

#read structure
my $mist = <DB>;
my $mist = <DB>;
$stru = <DB>;
chomp $stru;
#print "$stru\n";

# the pair tables of FILE1
my($i,$c,$hx,$px, @sslist, @pklist, @sstable, @pktable);
$hx=$i=0;
foreach $c (split(//,$stru)) {
  if ($c eq '(') {
    $sslist[++$hx]=$i;
  }
  elsif ($c eq ')') {
    if ($hx>0) {
      $sstable[$sslist[$hx]+1] = $i+1;
      $hx--;
    }
    else {
      substr($stru, $i, 1) = '>';
    }
  }
  elsif ($c eq '[') {
    $pklist[++$px]=$i;
  }
  elsif ($c eq ']') {
    if ($px>0) {
      $pktable[$pklist[$px]+1] = $i+1;
      $px--;
    }
    else {
      substr($stru, $i, 1) = '}';
    }
  }
  $i++;
}

#check whether the basepair is compatible with the sequence
for(my $i =0; $i<$len; $i++) {
  if ($sstable[$i]) {
    my $bi = $alnseq[$i-1];
    my $bj = $alnseq[$sstable[$i]-1];
    my $pair = $bi.$bj;
    my $corr = 0;

    foreach (@allow){
      $corr = 1 if $pair eq $_;
    }
    $sstable[$i] = 0  if !$corr;
  }
   if ($pktable[$i]) {
    my $bi = $alnseq[$i-1];
    my $bj = $alnseq[$pktable[$i]-1];
    my $pair = $bi.$bj;
    my $corr = 0;

    foreach (@allow){
      $corr = 1 if $pair eq $_;
    }
    $pktable[$i] = 0  if !$corr;
  }
}



my $i = 1;
my $s = 1;
foreach (@alnseq) {
  $number[$s] = $i++ if ($_ ne '-') ;
  $number[$s] = 0 if ($_ eq '-') ;
  $s++;
}

my @out;
for(my $i =0; $i<$seqlen; $i++) {
  $out[$i] = '.';
}

for(my $i =0; $i<$len; $i++) {
  if ($sstable[$i]) {
    $out[$number[$i]-1] = '(';
    $out[$number[$sstable[$i]]-1] = ')';
  }
  if ($pktable[$i]) {
  $out[$number[$i]-1] = '[';
  $out[$number[$pktable[$i]]-1] = ']';
  }
}

my $fertig;
for(my $i =0; $i<$seqlen; $i++) {
  $fertig .= $out[$i];
}

print $name,  "_out\n";
print "$seq \n";
print "$fertig \n";
