#!/usr/local/bin/perl
# -*-Perl-*-
# Last changed Time-stamp: <2004-08-18 12:58:43 xtina>

# reads one or two files containing dot-bracket structure ("mfe-format"
# including pseudoknots) ..((..[[..)).]]

# output: book-representation (linked diagramm) as ps-file
# different colors:
# black: basepairs contained in both input files
# red: basepairs contained only in the first file
# green: basepairs contained only in the second file

use Getopt::Std;
use strict;
use vars qw/$opt_b $opt_e $opt_o $opt_h  $N @file $hx $pf @outpairs
    $ss1 $ss2 $ssb $pk1 $pk2 $pkb/;

getopt('be');

usage() if($opt_h);

my $i=1;
while (my $fname=shift(@ARGV)) {  
  $file[$i++] = $fname ;
}

#open files
open(FILE1, "< $file[1]") || die("Can't open '$file[1]'!!!\n");
# If no second file is given, draw the first one only
(open(FILE2, "< $file[2]") || die("Can't open '$file[2]!!!\n")) if($file[2]);

#read 
my $name1 =<FILE1>;
my $seq1 = <FILE1>;
my $stru1 = <FILE1>;
chomp $name1;
chomp $seq1;
chomp $stru1;

my ($name2, $seq2, $stru2);
if($file[2]) {
  $name2 =<FILE2>;
  $seq2 = <FILE2>;
  $stru2 = <FILE2>;
  chomp $name2;
  chomp $seq2;
  chomp $stru2;
}
close(FILE1);
close(FILE2) if($file[2]);
  
$opt_b = 1 unless $opt_b;
$opt_e = length($seq1) unless $opt_e;

# take the part of the structure you want to view
$seq1 = substr($seq1, $opt_b-1, $opt_e-$opt_b+1);
$stru1 = substr($stru1, $opt_b-1, $opt_e-$opt_b+1);
$stru2 = substr($stru2, $opt_b-1, $opt_e-$opt_b+1) if($file[2]);
$N = length($stru1);

# replace non-matching brackets with '<', '>', '{' or '}' and make
# the pair tables of FILE1
my($i,$c,$hx,$px, @sslist1, @pklist1, @sstable1, @pktable1);
$hx=$i=0;
foreach $c (split(//,$stru1)) {
  if ($c eq '(') {
    $sslist1[++$hx]=$i;
  }
  elsif ($c eq ')') {
    if ($hx>0) {
      $sstable1[$sslist1[$hx]+1] = $i+1;
      $hx--;
    }
    else {
      substr($stru1, $i, 1) = '>';
    }
  }
  elsif ($c eq '[') {
    $pklist1[++$px]=$i;
  }
  elsif ($c eq ']') {
    if ($px>0) {
      $pktable1[$pklist1[$px]+1] = $i+1;
      $px--;
    }
    else {
      substr($stru1, $i, 1) = '}';
    }
  }
  $i++;
}
while ($hx>0) {
  $i = $sslist1[$hx--];
  substr($stru1, $i, 1) = '<';
}
while ($px>0) {
  $i = $pklist1[$px--];
  substr($stru1, $i, 1) = '{';
}

outpair($stru1, 1) if $opt_o;

# replace non-matching brackets with '<', '>', '{' or '}' and make
# the pair tables of FILE2
my (@sslist2, @pklist2, @sstable2, @pktable2,  @ssboth, @pkboth, $cf);
if($file[2]) {
  $hx=$i=0;
  foreach $c (split(//,$stru2)) {
    if ($c eq '(') {
      $sslist2[++$hx]=$i;
    }
    elsif ($c eq ')') {
      if ($hx>0) {
	$sstable2[$sslist2[$hx]+1] = $i+1;
	$hx--;
      }
      else {
	substr($stru2, $i, 1) = '>';
      }
    }
    elsif ($c eq '[') {
      $pklist2[++$px]=$i;
    }
    elsif ($c eq ']') {
      if ($px>0) {
	$pktable2[$pklist2[$px]+1] = $i+1;
	$px--;
      }
      else {
	substr($stru2, $i, 1) = '}';
      }
    }
    $i++;
  }
  while ($hx>0) {
    $i = $sslist2[$hx--];
    substr($stru2, $i, 1) = '<';
  }
  while ($px>0) {
    $i = $pklist2[$px--];
    substr($stru2, $i, 1) = '{';
  }
  
  outpair($stru2, 2) if $opt_o;

# for(my $i=1; $i<=$N; $i++) {
#    printf "%d %d 0\n", $i, $sstable1[$i] if $sstable1[$i];
#    printf "%d %d 1\n", $i, $pktable1[$i] if $pktable1[$i];
#  }
  


# find conflicting bases STIMMT SO NICHT 
#  $cf = 0;
#  my (@tab1, @tab2);
#  for(my $i=1; $i<=$N; $i++) {
#    ($tab1[$i] = $sstable1[$i]) && ($tab1[$sstable1[$i]] = $i)
#      if ($sstable1[$i]);
#    ($tab1[$i] = $pktable1[$i]) && ($tab1[$sstable1[$i]] = $i)
#	if ($pktable1[$i]);
#    ($tab2[$i] = $sstable2[$i]) && ($tab2[$sstable2[$i]] = $i)
#	if ($sstable2[$i]);
#    ($tab2[$i] = $pktable2[$i]) && ($tab2[$sstable2[$i]] = $i)
#	if ($pktable2[$i]);   
#  }
  
#  for(my $i=1; $i<=$N; $i++) {
#    $cf++ if (($tab1[$i]) && ($tab2[$i]) && ($tab1[$i] != $tab2[$i]));
#  }

 

  
# find common basepairs  
  $N = length($stru1);
  for(my $i=1; $i<=$N; $i++) {
    if($sstable1[$i] == $sstable2[$i]) {
      $ssboth[$i] = $sstable2[$i];
      $sstable1[$i] = 0;
      $sstable2[$i] = 0;
    }
    elsif ($pktable1[$i] == $sstable2[$i]) {
      $ssboth[$i] = $sstable2[$i];
      $pktable1[$i] = 0;
      $sstable2[$i] = 0;
    }
  }    

  for(my $i=1; $i<=$N; $i++) {
    if($pktable1[$i] == $pktable2[$i]) {
      $pkboth[$i] = $pktable2[$i];
      $pktable1[$i] = 0;
      $pktable2[$i] = 0;
    }
    elsif($sstable1[$i] == $pktable2[$i]) {
      $pkboth[$i] = $pktable2[$i];
      $sstable1[$i] = 0;
      $pktable2[$i] = 0;
    }
  }
  
}

# variables needed for ps
my $y = $N/2;
my $sc = $N+1;
my $t1 = 0;
my $t2 = 0;
my $colp1 = 'BLACK';
$colp1 = 'RED' if ($file[2]);
# write the postscript-file
my $ps;
if(($opt_b == 1) && ($opt_e == $N)) {
    $ps = $name1.'.ps';
  }
else {
    $ps = $name1.$opt_b.'_'.$opt_e.'.ps';
  }
open(PS, "> $ps") || die("Can't write ps-file!!!\n");

printf(PS "%%!PS-Adobe-2.0 EPSF-1.2
%%%%Title: $ps
%%%%Creator: comp_draw.pl
%%%%BoundingBox: 66 311 518 562
%%%%EndComments\n\n");

printf(PS "%%%%BeginProlog
/bookdict 100 dict def
bookdict begin
/fsize  {1} def
%% toggles: if you set them all to  false  the plot stays empty
/drawbases   true def  %% set to  false  to leave out sequence
/drawoutline false def  %% set to  false  to leave out backbone
/drawpairs   true def  %% set to  false  to not draw lines connecting pairs

%% colors
/BLACK { 0 0 0 } def
/RED   { 1 0 0 } def
/GREEN { 0 1 0 } def
/BLUE  { 0 0 1 } def
/LBLUE  { 0 1 1 } def
/YELLOW { 1 1 0 } def
/MAGENTA { 1 0 1 } def
/GREY {0.8 0.8 0.8} def
/ORANGE { 1 0.5 0 } def

%% Vars
/Xmax {$N } bind def
/Y { Xmax 2 div } bind def

%%Ellipse
/ellipsedict 8 dict def
ellipsedict /mtrx matrix put

/ellipse
{ ellipsedict begin
/endangle exch def
/startangle exch def
/yrad exch def
/xrad exch def
/y exch def
/x exch def

/savematrix mtrx currentmatrix def
x y translate
xrad yrad scale
0 0 1 startangle endangle arc
savematrix setmatrix
end
}def
    
%%procedures

/cshow { dup stringwidth pop fsize neg 3 div exch neg 2 div exch
         rmoveto show
       } bind def
/OUTLINE {
   gsave
     /Helvetica findfont fsize 3 mul scalefont setfont
    newpath
    1 Y moveto Xmax Y lineto
    1 1 Xmax { Y moveto (.) cshow } for
    stroke
    grestore
} bind def
/BASES {
   gsave
    /Helvetica findfont fsize scalefont setfont
    1 1 Xmax {
	dup Y moveto sequence exch 1 getinterval cshow
    } for
    grestore
} bind def
/NSSPAIRS {
    gsave
    setrgbcolor
    {newpath aload pop dup 3 -1 roll dup 3 1 roll add 2 div 3 1 roll sub 2 div dup 2 div Y 3 1 roll 360 180 ellipse stroke} forall
    grestore
} bind def

/SSPAIRS {
    gsave
    setrgbcolor
    {newpath aload pop dup 3 -1 roll dup 3 1 roll add 2 div 3 1 roll sub 2 div dup 2 div Y 3 1 roll 360 180 ellipse stroke} forall
    grestore
} bind def   
/PKPAIRS{
    gsave
    setrgbcolor      
    {newpath aload pop dup 3 -1 roll dup 3 1 roll add 2 div 3 1 roll sub 2 div dup 2 div Y 3 1 roll 180 360 ellipse stroke} forall
    grestore	
  } bind def
/OUTP{
  gsave 
  {newpath aload pop setrgbcolor arc stroke}forall
  grestore		
  } bind def	
    
");

printf(PS  "%% data start here
/sequence { (\n\\");
my @s1 = split(//, $seq1);
for(my $i=0; $i<$N; $i++) {
print PS $s1[$i];
printf PS "\\\n\\" if (!(($i+1)%78));
}
printf (PS "
) } def\n");


# the ss-edges only in the first file
$ss1 = 0;
printf(PS  "/sspairs1 [\n");
for(my $i=1; $i<=$N; $i++) {
  my $j = $sstable1[$i];
  printf(PS "[%d %d]\n", $i, $j ) && $ss1++
      if $j;
}
printf(PS  "]def\n");


# the ss-edges only in the second file
$ss2 = 0;
if($file[2]) {
  printf(PS  "/sspairs2 [\n");
  for(my $i=1; $i<=$N; $i++) {
    my $j = $sstable2[$i];
    printf(PS "[%d %d]\n", $i, $j) && $ss2++
if $j;
  }
  printf(PS  "]def\n");


# the ss-edges common to both files
$ssb = 0;
  printf(PS  "/ssboth [\n");
  for(my $i=1; $i<=$N; $i++) {
    my $j = $ssboth[$i];
    printf(PS "[%d %d]\n", $i, $j) && $ssb++
        if $j;
  }
  printf(PS  "]def\n");
}
# the pk-edges only in the first file
$pk1 = 0;
printf(PS  "/pkpairs1 [\n");
for(my $i=1; $i<=length($stru1); $i++) {
  my $j = $pktable1[$i];
  printf(PS "[%d %d]\n", $i, $j) && $pk1++
      if $j;
}
printf(PS  "]def\n");

# the pk-edges only in the second file
$pk2 = 0;
if($file[2]) {
  printf(PS  "/pkpairs2 [\n");
  for(my $i=1; $i<=length($stru1); $i++) {
    my $j = $pktable2[$i];
    printf(PS "[%d %d]\n", $i, $j) && $pk2++
        if $j;
  }
  printf(PS  "]def\n");

# the pk-edges common to both files
$pkb = 0;
  printf(PS  "/pkboth [\n");
  for(my $i=1; $i<=length($stru1); $i++) {
    my $j = $pkboth[$i];
    printf(PS "[%d %d]\n", $i, $j) && $pkb++
        if $j;
  }
  printf(PS  "]def\n");
}

# the pairs that go outside
if($opt_o) {
 printf(PS  "/outpaired [\n");
foreach(@outpairs) {
printf PS "[";
    for(my $j =0; $j<6; $j++){
	printf PS "$$_[$j] ";
      }
    printf PS "]\n";
  }
 printf(PS  "]def\n");
}

printf(PS  "end
%%EndProlog
");


  printf(PS  "
bookdict begin
1 setlinejoin
1 setlinecap
0.4 setlinewidth
72 216 translate
72 6 mul %f div dup scale\n", $sc);
  printf(PS  "/Helvetica findfont fsize scalefont setfont\n");
  printf(PS  "
%% draw the outline
drawoutline {
 OUTLINE
} if
    
%% draw bases
drawbases {
  BASES
} if 
    
%% draw base pairs
drawpairs {");
if($file[2]) {
printf(PS " 
ssboth   BLACK SSPAIRS
sspairs1 RED NSSPAIRS
sspairs2 GREEN   SSPAIRS
pkboth   BLACK PKPAIRS
pkpairs1 RED PKPAIRS
pkpairs2 GREEN   PKPAIRS
");
}
else {
printf(PS "
 sspairs1 BLACK SSPAIRS
 pkpairs1 BLACK PKPAIRS
");
}
printf(PS "
 outpaired OUTP
") if ($opt_o);
printf(PS "
} if
");	




    
 printf(PS  "  
%% show it
showpage
end
%%%%EOF\n");

close(PS);

printf "ss1 = %d, ss2 = %d, ssb = %d\npk1 = %d, pk2 = %d, pkb = %d\n",
$ss1, $ss2, $ssb, $pk1, $pk2, $pkb;

printf "reference structure: %d sspairs, %d pkpairs => %d pairs
hx structure:  %d pairs
%d correct pairs (%.1f %%), %d incorrect pairs \n",
$ss2+$ssb, $pk2+$pkb, $ss2+$ssb+$pk2+$pkb,
$ss1+$ssb+$pk1+$pkb,
$ssb+$pkb, ($ssb+$pkb)/($ss2+$ssb+$pk2+$pkb)*100, $ss1+$pk1 if $file[2];

printf "%d & %d & %d & %d & %.1f & %.1f \n", $ss2+$ssb+$pk2+$pkb,
$ssb+$pkb, $ss2+$pk2, $ss1+$pk1,  ($ssb+$pkb)/($ss2+$ssb+$pk2+$pkb)*100,
($ssb+$pkb)*100/($ssb+$pkb+$ss1+$pk1) if $file[2]; 

printf "\nSS=%.1f\nSP=%.1f\n\n", ($ssb+$pkb)/($ss2+$ssb+$pk2+$pkb)*100,
($ssb+$pkb)*100/($ssb+$pkb+$ss1+$pk1) if $file[2]; 



#**************************
sub outpair {
  my $stru = $_[0];
  my $col = $_[1];
  print $stru, "\n", $col, "\n";
  my @s = split //, $stru;
  my ($color, $rad);
  (($color = 'RED') && ($rad = $N/2)) if ($col eq 1);
  (($color = 'GREEN') && ($rad = $N/2.5)) if ($col eq 2);
  for(my $i =0; $i<$N; $i++) {
  SWITCH: {
   (push @outpairs, [$i+$rad, $N/2, $rad, 170, 180, $color]), last SWITCH if ($s[$i] eq '<');
   (push @outpairs, [$i+$rad, $N/2, $rad, 180, 190, $color]), last SWITCH if ($s[$i] eq '{');
   (push @outpairs, [$i-$rad, $N/2, $rad, 360, 10, $color]), last SWITCH if ($s[$i] eq '>');
   (push @outpairs, [$i-$rad, $N/2, $rad, 350, 360, $color]), last SWITCH if ($s[$i] eq '}');
  }
  }
  
  
}
#**************************
sub usage() {
  die "Usage: $0 [-b from] [-e to] file1 file2\n";
}
#**************************
