#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "xutils.h"
#include "hxmatch.h"
#include "read_graph.h"

INFO *alif_pairs;
int K;
void find_stacks(void);

INFO *alif(FILE *file) {

  char *line;
  int abase, atop, ainc, l;
  char nix[30];
  int bi, bj;
  int k=0;
  
  abase = 200;
  ainc = 100;
  atop = abase;
  
  if(!(alif_pairs = (INFO *)calloc(abase, sizeof(INFO)))) {
    fatal("write_alif_pairs(): allocation of alif_pairs failed!!!");
  }
  
  
  while((line = getline(file))) {  
    if(strcmp(line, "/pairs [") == 0)
      break;
     free(line);  
  }
  while((line = getline(file))) {
    if(strcmp(line, "] def") == 0)
      break;
    sscanf(line,"%c%d %d", nix, &bi, &bj);
    if(k>=atop) {
      if(!(alif_pairs = (INFO *)realloc(alif_pairs, (atop+ainc)*sizeof(INFO)))) {
	fatal("write_alif_pairs(): reallocation of alif_pairs failed!!!");
      }
      for(l=atop; l<atop+ainc; l++) {
	alif_pairs[l].bi =0;
	alif_pairs[l].bj =0;
	alif_pairs[l].xx =0;
      }
      atop += ainc;
    }
    alif_pairs[k].bi = bi;
    alif_pairs[k].bj = bj;
    alif_pairs[k++].wt = 4*AA;
    free(line); 
  }

  K = k;  
  find_stacks();

  

  return alif_pairs;
}
/* -----------------------------------------------------------*/
INFO *alp(FILE *file) {

  char *line;
  int abase, atop, ainc, l;
  char nix[30];
  char pr[10];
  int bi, bj;
  float p;
  float p_min;
  int k=0;

  p_min =90;
  abase = 200;
  ainc = 100;
  atop = abase;
  
  if(!(alif_pairs = (INFO *)calloc(abase, sizeof(INFO)))) {
    fatal("write_alif_pairs(): allocation of alif_pairs failed!!!");
  }

  line = getline(file);
  free(line);  
  line = getline(file);
  if(!(strcmp(line, "alifold output") == 0))
    fatal("The file given is not an 'alifold.out'-file\n");
  free(line);
  
  
  while((line = getline(file))) {
    sscanf(line,"%d %d %s %s", &bi,  &bj, nix, pr);
    p = atof(pr);

    if(k>=atop) {
      if(!(alif_pairs = (INFO *)realloc(alif_pairs, (atop+ainc)*sizeof(INFO)))) {
	fatal("write_alif_pairs(): reallocation of alif_pairs failed!!!");
      }
      for(l=atop; l<atop+ainc; l++) {
	alif_pairs[l].bi =0;
	alif_pairs[l].bj =0;
	alif_pairs[l].xx =0;
      }
      atop += ainc;
    }
    if(p>p_min) {
      alif_pairs[k].bi = bi;
      alif_pairs[k].bj = bj;
      alif_pairs[k++].wt = 8*AA*log(p/p_min);
    }
    free(line); 
  }

  K = k;  
  find_stacks();

  

  return alif_pairs;
}
/* -----------------------------------------------------------*/
INFO *ext_pairs(FILE *file) {

  char *line;
  int abase, atop, ainc, l;
  int bi, bj, wt;
  int k=0;
  
  abase = 200;
  ainc = 100;
  atop = abase;

 
  
  if(!(alif_pairs = (INFO *)calloc(abase, sizeof(INFO)))) {
    fatal("write_alif_pairs(): allocation of alif_pairs failed!!!");
  }
   
  while((line = getline(file))) {    
    sscanf(line,"%d %d %d", &bi, &bj, &wt);
   
    if(k>=atop) {
      if(!(alif_pairs = (INFO *)realloc(alif_pairs, (atop+ainc)*sizeof(INFO)))) {
	fatal("write_alif_pairs(): reallocation of alif_pairs failed!!!");
      }
      for(l=atop; l<atop+ainc; l++) {
	alif_pairs[l].bi =0;
	alif_pairs[l].bj =0;
	alif_pairs[l].xx =0;
      }
      atop += ainc;
    }
    alif_pairs[k].bi = bi;
    alif_pairs[k].bj = bj;
    alif_pairs[k].xx = 1;
    alif_pairs[k++].wt = 4*wt;
    
    free(line); 
  }

  K = k;  

  
  return alif_pairs;
}
/* -----------------------------------------------------------*/
void find_stacks(void) {
  
  /* skip isolated basepairs (xx = 0), mark the beginnining(xx = 2) and
     end of stacks (xx = 3) */
  
  int bst, est; /* to mark the beginning resp. end of a stack */
  int i, isbp;
  
  isbp = 0;
  for(i = 1; i < K-1; i++) {
    alif_pairs[i].xx = 1;
  }
  /* are the first or the last basepair in the list isolated ? */
  if(!((alif_pairs[1].bi == alif_pairs[0].bi+1) &&
       (alif_pairs[1].bj == alif_pairs[0].bj-1))) {
    alif_pairs[0].xx = 0;
    isbp++;
  }
  else alif_pairs[0].xx = 2;  
  if(!((alif_pairs[K-2].bi == alif_pairs[K-1].bi-1) &&
       (alif_pairs[K-2].bj == alif_pairs[K-1].bj+1))) {
    alif_pairs[K-1].xx = 0;
    isbp++;
  }
  else alif_pairs[K-1].xx = 3;
  
  for(i = 1; i < K-1; i++) {
    bst = est = 0;
    /* is the basepair the beginning of a stack? */
    if(!((alif_pairs[i-1].bi == alif_pairs[i].bi-1) &&
	 (alif_pairs[i-1].bj == alif_pairs[i].bj+1))) {
      alif_pairs[i].xx = 2;
      bst =1;
    }
    /* is the basepair the end of a stack? */
    if( !((alif_pairs[i+1].bi == alif_pairs[i].bi+1) &&
	  (alif_pairs[i+1].bj == alif_pairs[i].bj-1))) {
      alif_pairs[i].xx = 3;
      est = 1;
    }
    /* if the basepair is isolated, set xx = 0 */
    if(bst && est) {
      alif_pairs[i].xx = 0;
      isbp++;
    }
  }
  
/* set .xx = 0 if stacklen<3 */
  for(i = 0; i < K; i++) {
    if((alif_pairs[i].xx ==2) && (alif_pairs[i+1].xx == 3) ) {
      alif_pairs[i].xx = 0;
      alif_pairs[i+1].xx = 0;
      isbp += 2;
    }
  }
   
}

/*************************************************************************/
INFO *mix_bpairs(INFO *alif_bp, INFO *bpairs) {

  int i, j, ii, jj, max_wt;

  max_wt = 0;
  /* find maximum weight */
  for(i = 0; i < Nbp; i++) {
    if(bpairs[i].wt > max_wt)
      max_wt = bpairs[i].wt;
  }
  
  for(i=0; i<K; i++) {
    if(alif_bp[i].xx) {
      ii = alif_bp[i].bi;
      jj = alif_bp[i].bj;
      for(j=0; j<Nbp; j++) {
	if((bpairs[j].bi == ii) && (bpairs[j].bj == jj)) {
	  bpairs[j].wt += alif_bp[i].wt;
	  bpairs[j].owt += 50; 
	  continue;
	}
      }
    }
  }


  return bpairs;
}
