/* Last changed Time-stamp: <2004-09-03 13:49:09 xtina> */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <math.h>
#include "xutils.h"
#include "hx_pair_math.h"
#include "hxmatch.h"
#include "read_graph.h"
#include "list.h"

/* #include "/usr/local/include/dmalloc.h"  */

#define LOOP_MIN 3 /* minimal loopsize for hairpin*/
#define H_BASE 100 /* base size of H_array */
#define H_INC 100
#define P_BASE 1000 /* base size of bpairs */


typedef struct _helix {
  short len;
  short bp[2]; /* bp[0]=i; bp[1]=j; with i<j */
  double energy;
  short mut;
  list *pairs;
} HELIX;

typedef struct _op {
  int pair;
  int helix;
} OP;


static void make_matrix (void);
static short *encode_sequence(char *seq,int **aln_ct);
static int generate_helices (short *enc_seq);
static int HelixLength(int nii, int nff, short *enc_seq); 
static int cmp_infwt(const void *w1, const void *w2);
static void calc_energy(short *enc_seq);
static BASINF *find_bpairs(int seq_nbp);
static void bp_together(int seq_nbp, int seq_nr);
static void weight_pairs(void);
static void ini_bp(void);
static void aln_number(int seq_nbp, int *aln_count);
static INFO *pen_lr(INFO *bpairs);
static int cmp_bi(const void *i1, const void *i2);
static int cmp_bj(const void *j1, const void *j2);
int new_helix(int last, int hx);
int normal(int last, int hx);
int left_bulge(int last, int hx, int regular);
int right_bulge(int last, int hx, int regular);
OP *new_op(int pair, int hx);
void copy_helix(int hx, int regular);
int exist_npw(int last);
int *new_it(int idx);

static int H_len;
static int H_count;
static HELIX *H_array;
static INFO *bpairs;
static BASINF *seq_bpairs;
static BASINF *aln_bpairs;
static int *indx;
static int *Index;
static INFO **Pindex;
static int max_gap;
static double sc_fact, nc_fact;
static int hx_size, hx_inc;
static HELIX *aln_hx;

INFO *helix(int h_len) {

  int i;
  
  /* set the parameters */
  sc_fact =  100*phi2;
  nc_fact = phi1;
  Min_comp =  (int) (Nseq/2); 
  max_gap =  (int) (Nseq/2);
  /*initialize */
  H_len = h_len;
  V = strlen(Aln_seqs[1]);
  ini_bp();

  fprintf(stderr, "calculating thermodynamic score ...\n");
  make_matrix();

  fprintf(stderr, "calculating covariation score ...\n");
  weight_pairs();

  fprintf(stderr, "scanning the combined scoring matrix for helices ...\n");

  bpair2helix(bpairs);

  bpairs = pen_lr(bpairs);
 
 /*  bpairs = choose_bp(bpairs); */
   /* weight is an even integer (and rescaled)*/
  for(i = 0; i <  Nbp; i++) {
    bpairs[i].wt = 4*(int)((float)bpairs[i].wt/100);
  }

  qsort(bpairs, Nbp, sizeof(INFO), cmp_bj);
  qsort(bpairs, Nbp, sizeof(INFO), cmp_bi);

#if 0
  for(i=0;i<Nbp;i++) {
    if(bpairs[i].wt>0)
      printf("%d %d %d %d %d\n", bpairs[i].bi, bpairs[i].bj,
	     bpairs[i].wt, bpairs[i].mut, bpairs[i].comp); 
  }
#endif  
  return bpairs;
}

/***************************************************************************/
INFO *pen_lr(INFO *bpairs){
  int i, diff;
  
  for(i = 0; i <  Nbp; i++) {
    diff = bpairs[i].bj - bpairs[i].bi;
    if( (diff > 400) && (diff <= 800))
      bpairs[i].wt -= 40*(diff-400);
    if(diff > 800)
      bpairs[i].wt -= 16000;     
  } 
  return bpairs;
}

/***************************************************************************/
static void make_matrix (void) {

  char *sequence;
  short *enc_seq;
  int *aln_count;
  int i, seq_nbp;

  for(i=1; i<=Nseq; i++) { /* for each sequence ... */
    fprintf(stderr,"sequence %d\n", i);
    if(!(H_array = (HELIX *)calloc(H_BASE, sizeof(HELIX))))
      fatal("helix(): allocation of H_array failed!!!");
    
    sequence = Aln_seqs[i];
    
    /* encode_sequence() also removes gaps;
       aln_count keeps track of the aln-numbering;
       the following functions work without gaps; */
    enc_seq = encode_sequence(sequence, &aln_count);
    
    seq_nbp = generate_helices(enc_seq);    
    
    calc_energy(enc_seq);

    seq_bpairs = find_bpairs(seq_nbp); /* H_array to seq_bpairs*/

    free(H_array);
    H_array = NULL;
    free(enc_seq);
    enc_seq = NULL;
    
    /* changes the numbering of the positions back to aln-numbering */
    aln_number(seq_nbp, aln_count);

     /* free */
    free(seq_bpairs);
    seq_bpairs = NULL;
    free(aln_count);
    aln_count = NULL;
    
    bp_together(seq_nbp, i);

    free(aln_bpairs);
    aln_bpairs = NULL;
  }
  free(Index);
  Index = NULL;
  free(indx);
  indx = NULL;
}

/***************************************************************************/
static short *encode_sequence(char *seq, int **aln_ct) {

  int i, ii, act;
  char *ngseq;
  char alphi[] = "_ACGUXKIN"; /* A: 1, C: 2, G: 3, U: 4, X: 5, K: 6, I: 7 N: 8 */
  char *tmp;
  short *s;
  int *aln_count;

  /* aln_count */
  if (!(aln_count = (int *)calloc((V+1), sizeof(int)))) {
    fatal("encode_sequence(): allocation of a_c failed!!!");
  }
  /* no gap sequence */
  if (!(ngseq = (char *)calloc((V+1), sizeof(char)))) {
    fatal("encode_sequence(): allocation of ngseq failed!!!");
  }
  act = 0; 

  for (i = 0; i<V; i++) {
    if (seq[i] == '-') continue;
    else {
      ngseq[act] = seq[i];
      /*alncount keeps for each nt i the aln numbering in aln[i] */
      aln_count[act+1] = i+1; 
      act++;
    }
  }

  for (i = act; i<V; i++) {
    ngseq[i] = '-';
    /* unused places of alncount are marked with a negative number */
    aln_count[i+1] = -i-1; 
  }

  if (!(s = (short *)calloc((V+1), sizeof(short)))) {
    fatal("encode_sequence(): allocation of s failed!!!");
  }
  for (i = 0; i<V; i++) {
    ngseq[i]=toupper(ngseq[i]);
    if ((tmp=strchr(alphi, ngseq[i]))) ii = tmp-alphi;
    else ii = 0;
    s[i] = ii;
  }
  free(ngseq);
  
  *aln_ct = aln_count;

  return s;
}

/***************************************************************************/
/* generate the longest stack a bp is part of */
static int generate_helices (short *enc_seq) {

  int H_top;
  int i, j, l;
  int len, seq_nbp;
  int cons = 2*H_len+LOOP_MIN-1;
  
  H_count = 0;
  H_top = H_BASE;
  seq_nbp = 0;
  for (i=0; i <= V-cons; i++) {
    for(j=V-1; j >= i+cons; j--) {
      /* i,j can't pair */
      if (!PairMx[enc_seq[i]][enc_seq[j]]) continue;
      /* i,j are not the beginning of a helix */
      if (i>0 && j < V-1)
	if (PairMx[enc_seq[i-1]][enc_seq[j+1]]) continue;
      len = HelixLength(i, j, enc_seq);
      if (len >= H_len) { 
	if (H_count >= H_top) {
	  if (!(H_array = (HELIX *)realloc(H_array,
					   (H_top+H_INC)*sizeof(HELIX))))
	    fatal("HelForm(): realloc of H_array failed!!!");
	  for(l=H_top; l<H_top+ H_INC; l++) {
	      H_array[l].len =0;
	      H_array[l].bp[0] =0;
	      H_array[l].bp[1] =0;
	      H_array[l].energy =0;
	      H_array[l].mut =0;	      
	    }
	  H_top += H_INC;
	}
	H_array[H_count].len = len;
	H_array[H_count].bp[0] = i;
	H_array[H_count++].bp[1] = j;
	seq_nbp += len;
      }
    }
  }
  return seq_nbp;
}

/***************************************************************************/
static int HelixLength(int nii, int nff, short *enc_seq) {

  int len=1;

  for(;;) {
    /* minloop condition is violated */
    if(nff-nii-2*len <= LOOP_MIN) break;
    /* end of stack */
    if(!PairMx[enc_seq[nii+len]][enc_seq[nff-len]]) break;
    len++;
  }
  return(len);
}

/***************************************************************************/
static int cmp_bi(const void *i1, const void *i2) {
  return ((INFO *)i1)->bi - ((INFO *)i2)->bi;
}

/***************************************************************************/
static int cmp_bj(const void *j1, const void *j2) {
  return ((INFO *)j1)->bj - ((INFO *)j2)->bj;
}

/***************************************************************************/
static int cmp_infwt(const void *w1, const void *w2) {
  return ((INFO *)w2)->wt - ((INFO *)w1)->wt;
}

/***************************************************************************/
/* calculate the energy of each helix with the help of hx_pair_math */
static void calc_energy(short *enc_seq) {
  int i, k, ii, jj, len;
  int bp1, bp2;
  double H = 0;
  double G = 0;
  double energy;
  double tempf;
  
  tempf = (temp +K0)/Tmeasure;
  for(i = 0; i < H_count; i++) {
    energy = 0;
    ii = H_array[i].bp[0];
    jj = H_array[i].bp[1];
    len = H_array[i].len;
    /* start with innermost basepair */
    for(k = len - 1; k > 0; k--) {
      bp1 = PairMx[enc_seq[ii+k]][enc_seq[jj-k]];
      bp2 = PairMx[enc_seq[ii+k-1]][enc_seq[jj-k+1]];
      H = enthalpies[bp1][bp2];
      G = stack37[bp1][bp2];
      energy += (H - (H - G)*tempf);
    }
     H_array[i].energy = energy;  
  }
}

/* bpairs lists all possible basepairs (including those with positive energy =
   negative weight), each basepair ocurres only once */
/* the weight of each basepair is the energy of the corresponding helix */

static BASINF *find_bpairs(int seq_nbp) {
  int i, j, k, len;
  double e;
  BASINF *tpairs; /* temporary bpairs */
 
  if(!(tpairs = (BASINF *)calloc(seq_nbp+1, sizeof(BASINF)))) {
    fatal("write_bpairs(): allocation of tpairs failed!!!");
  }
  
  j = 1;   
  for (i = 0; i < H_count; i++) {
    len = H_array[i].len;
    for(k = 0; k < len; k++) {
      /* start counting the bases with 1 instead of 0
	 (necessary for wmatch) */
      tpairs[j].bi = H_array[i].bp[0] + k + 1;
      tpairs[j].bj = H_array[i].bp[1] - k + 1;
      e =  H_array[i].energy;
      /* wt is energy of helix, positive and an integer */
      tpairs[j].wt = -(int)( e);  
      /* tpairs[j].wt = 100*H_array[i].len;   */
      j++;
    }
  }

  return tpairs;  
}

/***************************************************************************/
static void aln_number(int seq_nbp, int *aln_count) {
  
  int i;
  int ii, jj;
    
  if(!(aln_bpairs = (BASINF *)calloc(seq_nbp+1, sizeof(BASINF)))) {
    fatal("aln_number(): allocation of aln_bpairs failed!!!");      
  }
  
  for(i=1; i<=seq_nbp; i++) {
    ii = seq_bpairs[i].bi;
    jj = seq_bpairs[i].bj;
    aln_bpairs[i].bi = aln_count[ii];
    aln_bpairs[i].bj = aln_count[jj];
    aln_bpairs[i].wt = seq_bpairs[i].wt;
  }
}

/***************************************************************************/
void bp_together (int seq_nbp, int seq_nr) {
  int i;
  int ii, jj, idx;
  
  static int ptop;
  int ALC;
  ALC = 10000000;
  /* make an entry in Index for each new pair */
  /* only when the seq_nr <= Nseq - Min_comp+1! */

  if(seq_nr <= Nseq - Min_comp + 1) {
    for (i=1; i<= seq_nbp; i++) {
      ii = aln_bpairs[i].bi;
      jj = aln_bpairs[i].bj;
      if (Index[indx[jj]+ii] == 0) {
	Index[indx[jj]+ii] = ++Nbp;	
      }
    }
    /* reallocate bpairs */
    if(Nbp > ptop) {
      if(!(bpairs = (INFO *)realloc(bpairs, Nbp*sizeof(INFO)) )) {
	fatal("bp_together(): reallocation of bpairs failed!!!");
      }
      cset_info(bpairs, ptop, Nbp); /* set the allocated memory to zero */
      ptop = Nbp;
    }
  }
  
  /* make an entry in bpairs */
  for (i=1; i<= seq_nbp; i++) { 
    ii = aln_bpairs[i].bi; 
    jj = aln_bpairs[i].bj; 
    idx = Index[indx[jj]+ii]; 
    if (idx == 0) continue;  /* skip this entry */
    bpairs[idx-1].bi = ii; /* start at position 0 of bpairs*/
    bpairs[idx-1].bj = jj;
    bpairs[idx-1].wt += aln_bpairs[i].wt;
  }

  return;
}

/**************************************************************************/  
void weight_pairs(void) {

  int i, j, ii, jj, k, l;
  int comp, mut, mwt, pbp;
  double score, wquer, sc1, sc2, q;
  char ci, cj;
  
  int dm[7][7]={{0,0,0,0,0,0,0}, /* hamming distance between pairs */
	       	{0,0,2,2,1,2,2} /* CG */,
		{0,2,0,1,2,2,2} /* GC */,
		{0,2,1,0,2,1,2} /* GU */,
		{0,1,2,2,0,2,1} /* UG */,
		{0,2,2,1,2,0,2} /* AU */,
		{0,2,2,2,1,2,0} /* UA */};
  
  /* scale the weight */ 
  for (i=0; i<Nbp; i++) {
    bpairs[i].wt = (int) (bpairs[i].wt/Nseq); 
  }
  
  /*calculate max and mean weight */
  mwt =0;
  wquer=0;
  pbp =0;  
  for (i=0; i<Nbp; i++) {
    if(mwt<bpairs[i].wt)
      mwt = bpairs[i].wt;
  }  
  for (i=0; i<Nbp; i++) {
    if( bpairs[i].wt>0) {
      wquer+=bpairs[i].wt;
      pbp++;
    }
  }
  wquer = wquer/pbp;
  
  /* have a look whether compatible or not */
  for (i=0; i<Nbp; i++) {
    ii = bpairs[i].bi;
    jj = bpairs[i].bj;
    
    for ( k = 1; k<= Nseq; k++) {
      ci = Aln_seqs[k][ii-1];
      cj = Aln_seqs[k][jj-1];
      bpairs[i].ptyp[PairSeq[ENCODE(ci)][ENCODE(cj)] ]++; 
    }
  }
  for (i=0; i<Nbp; i++) {
    comp = 0;
    for(j=1; j<=7; j++) { /* gap-gap  included */
      comp += bpairs[i].ptyp[j];
    }
    bpairs[i].comp = comp;
  }
  
  /* count compensatory mutations: */
  for (i=0; i<Nbp; i++) {
    mut = 0;
    for(j=1; j<7; j++) {
      if (bpairs[i].ptyp[j]) {
	mut++;
      }
    }
    bpairs[i].mut = mut;
    
    /* Weighting: */
    score =0;
    for(k=1; k<=6; k++) {   
      for (l=k+1; l<=6; l++) {
	/* scores for replacements between pairtypes    */
	/* consistent or compensatory mutations score 1 or 2  */
	score += bpairs[i].ptyp[k]*bpairs[i].ptyp[l]*dm[k][l];
      }
    }
    sc1 = score/((Nseq)*(Nseq));
    
    q = ((double)(bpairs[i].ptyp[0]))/(double)(Nseq);
    
    /* counter examples score */
    sc2 = sc1 - nc_fact*q;
    
    bpairs[i].wt += (int) (sc_fact * sc2);
    
    bpairs[i].owt = bpairs[i].wt;
  }
}
  
/***************************************************************************/
void bpair2helix(INFO *bpairs) {
 
  int i, l;
  int hxct, hx;
  int last, new, found, regular;
  int left, right;
  int *zz;
  int *pp;
  int max, calsize;
  int ii, ij;
  int LL, nh, mct, max_nh;
  float wt, mwt, mhxl;
  list *open;
  OP *oo;
  L_Item *li;
  
  hx_size = 3000;
  hx_inc = 3000;
  
  qsort(bpairs, Nbp, sizeof(INFO), cmp_bj);
  qsort(bpairs, Nbp, sizeof(INFO), cmp_bi);


  /****/
  if(!(indx =  (int *) calloc((V+1), sizeof(int)))) {
    fatal("bp_together(): allocation of indx failed!!!");
  }
  for (i = 1; i <= V; i++) {
    indx[i] = (i*(i-1))/2;
  }
  calsize = (V * (V+1) ) / 2;
  if( !(Pindex = (INFO **) calloc(calsize, sizeof(INFO *)) ) ) {
    fatal("bp_together(): allocation of Index failed!!!");
  }
  /****/
  
  /* find helices * */
  /*initialize*/
  if(!(aln_hx = (HELIX *)calloc(hx_size, sizeof(HELIX))))
     fatal("bpair2helix(): allocation of aln_hx  failed!!!");
  for(i=0; i<Nbp; i++) {
    bpairs[i].xx = -1;
    bpairs[i].helices = new_list();
    if (bpairs[i].wt<0)
      bpairs[i].wt = -500;
    ii = bpairs[i].bi;
    ij = bpairs[i].bj;
    Pindex[indx[ij]+ii] = &bpairs[i];
  }
 
  hxct = 0;
  open = new_list();
  
  for(i=0; i<Nbp; i++) {
    last = 0;
    if (bpairs[i].xx > 0) /* nur neue als beginn einer helix */
      continue;    
    if (bpairs[i].wt < 0) /* nur pos wt als beginn einer helix */
      continue;    
    
    hxct++;
    if(hxct >= hx_size) {
      hx_size += hx_inc;
      if(!(aln_hx = (HELIX *)realloc(aln_hx,
				     hx_size*sizeof(HELIX))))
	fatal("bpair2helix(): reallocation of aln_hx  failed!!!");
      for(l=hx_size-hx_inc; l<hx_size; l++) {
	aln_hx[l].len =0;
	aln_hx[l].bp[0] =0;
	aln_hx[l].bp[1] =0;
	aln_hx[l].energy =0;
	aln_hx[l].mut =0;	      
      }
    }
    last = new_helix(i, hxct); /*is there a helix beginning with basepair i? */

    if(!last) {
      hxct--;
      continue;
    }

    oo = new_op(last, hxct);
    l_insert(open, oo, L_TAIL);
    
    li = open->head;    
    while(li != NULL) {
      oo = (OP *)(li->item);
      last = oo->pair;
      hx = oo->helix;
      regular = hx;
      free(oo);
      l_remove(open, L_HEAD);
      
      
      new = 0;
      found =1;
      /* normal weiter */
      while(found != 0) {
	found =0;
	new = normal(last, hx);
	if (new != 0) {
	  last = new;
	  found =1;
	}
      }
       
      hxct++;
      if(hxct >= hx_size) {
	hx_size += hx_inc;
	if(!(aln_hx = (HELIX *)realloc(aln_hx,
				       hx_size*sizeof(HELIX))))
	  fatal("bpair2helix(): reallocation of aln_hx  failed!!!");
	for(l=hx_size-hx_inc; l<hx_size; l++) {
	  aln_hx[l].len =0;
	  aln_hx[l].bp[0] =0;
	  aln_hx[l].bp[1] =0;
	  aln_hx[l].energy =0;
	  aln_hx[l].mut =0;	      
	}
      }
      left = left_bulge(last, hxct, regular);
      if (left) {
	oo = new_op(left, hxct);
	l_insert(open, oo, L_TAIL);
	hxct++;
	if(hxct >= hx_size) {
	  hx_size += hx_inc;
	  if(!(aln_hx = (HELIX *)realloc(aln_hx,
					 hx_size*sizeof(HELIX))))
	    fatal("bpair2helix(): reallocation of aln_hx  failed!!!");
	  for(l=hx_size-hx_inc; l<hx_size; l++) {
	    aln_hx[l].len =0;
	    aln_hx[l].bp[0] =0;
	    aln_hx[l].bp[1] =0;
	    aln_hx[l].energy =0;
	    aln_hx[l].mut =0;	      
	  }
	}
      }         
      right = right_bulge(last, hxct, regular);
      if (right) {
	oo = new_op(right, hxct);
	l_insert(open, oo, L_TAIL);
        hxct++;
	if(hxct >= hx_size) {
	  hx_size += hx_inc;
	  if(!(aln_hx = (HELIX *)realloc(aln_hx,
					 hx_size*sizeof(HELIX))))
	    fatal("bpair2helix(): reallocation of aln_hx  failed!!!");
	  for(l=hx_size-hx_inc; l<hx_size; l++) {
	    aln_hx[l].len =0;
	    aln_hx[l].bp[0] =0;
	    aln_hx[l].bp[1] =0;
	    aln_hx[l].energy =0;
	    aln_hx[l].mut =0;	      
	  }
	}
      }      
      hxct--;
      
      li = open->head;
    }
  }



  
  /* the weight of a basepair is the weight of the helix */
  mwt = 0;
  LL = 0;
  nh = 0;
  max_nh = 0;
  
  for(i=0; i<Nbp; i++) {
    max=0;
    mct = 0;
    li = bpairs[i].helices->head;
    while (li != NULL){
      zz = (int *)(li->item);
      nh++;
      mct++;
      if(aln_hx[zz[0]].energy > max)
	max = aln_hx[zz[0]].energy;
      LL += aln_hx[zz[0]].len;
      li = li->nxt;
    }
    wt = max;
    if(mct>max_nh)
      max_nh = mct;
    bpairs[i].wt = (int)(wt); 
    if (wt > mwt)
      mwt = wt;
  }
  mhxl = ((float)(LL))/((float)(nh));

  /* clean up */  
  free(indx);
  free(Pindex);

  for(i=0; i<Nbp; i++) { 
    li = bpairs[i].helices->head;
    while (li != NULL){
      zz = (int *)(li->item);           
      if(zz) {	
	free(zz);
	zz = NULL;
      }
      li->item = NULL;
      li = li->nxt;
    }
    L_Destroy(bpairs[i].helices, NULL); 
    bpairs[i].helices = NULL; 
  }
 
  
  for(i=1; i<=hxct ;i++){
    li = aln_hx[i].pairs->head;
    while (li != NULL){
      pp = (int *)(li->item);
      if(pp) {	
	free(pp);
	pp = NULL;
      }
      li->item = NULL; 
      li = li->nxt;
    }
    L_Destroy(aln_hx[i].pairs, NULL);
  }

  L_Destroy(open, NULL);
  free(aln_hx);

  return;  
}

/*************************************************************************/
int *new_it(int idx) {
  int *ii;
  
  if(!(ii = (int *)calloc(1, sizeof(int))))
    fatal("new_it: allocation of ii  failed!!!");
  ii[0] = idx;

  return ii;
}

/*************************************************************************/
OP *new_op(int pair, int hx) {
  OP *oo;
  
  if(!(oo = (OP *)calloc(1, sizeof(OP))))
    fatal("new_op: allocation of oo  failed!!!");
  oo->pair = pair;
  oo->helix = hx;
  
  return oo;
}

/*************************************************************************/
int right_bulge(int last, int hx, int regular){
  int j, ii, ij, ji, jj;
  int right, new, *it;
  
  right = 0;
  ii = bpairs[last].bi;
  ij = bpairs[last].bj;

  
  if(Pindex[indx[ij-2]+ii+1]){ /* right bulge */
    j = (int) (Pindex[indx[ij-2]+ii+1]-bpairs);
    ji = bpairs[j].bi; 
    jj = bpairs[j].bj;
    
    new = exist_npw(j);
    if(new !=0) {
      right = j;
      /*copy helix*/
      copy_helix(hx, regular);
      /*bp j in helix*/
      it = new_it(j);	
      l_insert(aln_hx[hx].pairs, it, L_TAIL);
      it = new_it(hx);
      l_insert(bpairs[j].helices, it, L_TAIL);
      bpairs[j].xx = 1;
      aln_hx[hx].len++;
      aln_hx[hx].energy += bpairs[j].wt;
      aln_hx[hx].mut += bpairs[j].mut;
    }    	
  }
  
  return right;
}

/*************************************************************************/
int left_bulge(int last, int hx, int regular){
  int j, ii, ij, ji, jj;
  int left, new, *it;
  
  left = 0;
  ii = bpairs[last].bi;
  ij = bpairs[last].bj;


  if(Pindex[indx[ij-1]+ii+2]) { /* links bulge */
    j = (int) (Pindex[indx[ij-1]+ii+2]-bpairs);
    ji = bpairs[j].bi; 
    jj = bpairs[j].bj;
    new = exist_npw(j);
    if(new !=0) {
      left = j;
      /*copy helix*/
      copy_helix(hx, regular);
      /*bp j in helix*/
      it = new_it(j);	
      l_insert(aln_hx[hx].pairs, it, L_TAIL);
      it = new_it(hx);
      l_insert(bpairs[j].helices, it, L_TAIL);
      bpairs[j].xx = 1;
      aln_hx[hx].len++;
      aln_hx[hx].energy += bpairs[j].wt;
      aln_hx[hx].mut += bpairs[j].mut;
    }    
  }    
  return left;
}

/***************************************************************************/
void copy_helix(int hx, int regular) {
  int *pp, *it;
  L_Item *li;

  aln_hx[hx].energy = aln_hx[regular].energy;
  aln_hx[hx].len  = aln_hx[regular].len;
  aln_hx[hx].mut = aln_hx[regular].mut;
    
  aln_hx[hx].pairs = new_list();
  
  li = aln_hx[regular].pairs->head;
  while (li != NULL){ /* alte basenpaare in neue helix eintragen */
    pp = (int *)(li->item);
    it = new_it(pp[0]);
    l_insert(aln_hx[hx].pairs, it, L_TAIL);
    it = new_it(hx);
    l_append(bpairs[pp[0]].helices, it); /* neue helix in alte basenpaare eintragen */
    
    li = li->nxt;
  }
}

/***************************************************************************/
int exist_npw(int last) {
  int j, ii, ij;
  int exist;
  
  exist = 0;
  ii = bpairs[last].bi;
  ij = bpairs[last].bj;
 
  /* normale verlaengerung */
  if((Pindex[indx[ij-1]+ii+1])) { 
    j = (int) (Pindex[indx[ij-1]+ii+1]-bpairs);
    if (bpairs[j].wt >0)
      exist = j;    
  }
  
  return exist;
}

/***************************************************************************/
int normal(int last, int hx) {
  int j, ii, ij, ji, jj;
  int new, ex;
  int *it;
  
  new = 0;
  ii = bpairs[last].bi;
  ij = bpairs[last].bj;
  if(Pindex[indx[ij-1]+ii+1]){ /* normale verlaengerung */
    j = (int) (Pindex[indx[ij-1]+ii+1]-bpairs);
    ji = bpairs[j].bi; 
    jj = bpairs[j].bj;
    
    if(bpairs[j].wt < 0 ) {
      ex = exist_npw(j);
      if (ex == 0)
	return 0;
    }
    
    new = j;
    /*bp j in helix*/
    it = new_it(j);
    l_insert(aln_hx[hx].pairs, it, L_TAIL);
    it = new_it(hx);
    l_insert(bpairs[j].helices, it, L_TAIL);
    bpairs[j].xx = 1;
    aln_hx[hx].len++;
    aln_hx[hx].energy += bpairs[j].wt;
    aln_hx[hx].mut += bpairs[j].mut;
  }
  
  return new;
}

/***************************************************************************/
int new_helix(int last, int hx) {
  int j, ii, ij, ji, jj;
  int new, ex, *it;
  int l;
 
  
  new = 0;  
  ii = bpairs[last].bi;
  ij = bpairs[last].bj;


  if(Pindex[indx[ij-1]+ii+1]) {  /* normale verlaengerung */
    j = (int) (Pindex[indx[ij-1]+ii+1]-bpairs);
    ji = bpairs[j].bi; 
    jj = bpairs[j].bj;

    if(bpairs[j].wt < 0 ) {
      ex = exist_npw(j);
      if (ex == 0)
	return 0;
    }
    new = j;
    if(hx >= hx_size) {
      hx_size += hx_inc;
      if(!(aln_hx = (HELIX *)realloc(aln_hx,
				     hx_size*sizeof(HELIX))))
	fatal("bpair2helix(): reallocation of aln_hx  failed!!!");
      for(l=hx_size-hx_inc; l<hx_size; l++) {
	aln_hx[l].len =0;
	aln_hx[l].bp[0] =0;
	aln_hx[l].bp[1] =0;
	aln_hx[l].energy =0;
	aln_hx[l].mut =0;	      
      }
    }
    bpairs[last].xx = 2;
    aln_hx[hx].bp[0] = bpairs[last].bi;
    aln_hx[hx].bp[1] = bpairs[last].bj;
    aln_hx[hx].len = 1;
    aln_hx[hx].energy = bpairs[last].wt;	  
    aln_hx[hx].mut = bpairs[last].mut;
    
    aln_hx[hx].pairs = new_list();
    it = new_it(last);
    l_insert(aln_hx[hx].pairs, it, L_TAIL);
    it = new_it(j);
    l_insert(aln_hx[hx].pairs, it, L_TAIL);
    it = new_it(hx);
    l_insert(bpairs[last].helices, it, L_TAIL);
    it = new_it(hx);
    l_insert(bpairs[j].helices, it, L_TAIL);
    
    /* und j setzt sie fort */
    bpairs[j].xx = 1;
    aln_hx[hx].len++;
    aln_hx[hx].energy += bpairs[j].wt;
    aln_hx[hx].mut += bpairs[j].mut;        
    }
  return new;
}

/***************************************************************************/
void no_mwm(INFO *bpairs) {
 
  int i, j;
  int ii, ij, ji, jj;
  int mc, pk;

  qsort(bpairs, Nbp, sizeof(INFO), cmp_infwt);
  
  /* color the bp's (different colors are 0, 1, 2,... in outcom[i].col[0]) */
  pk = 0;
  mc = 1;
  while(mc) { /* if mc = 0 after a search, there are no more colors*/
    mc = 0;
    pk++; /* pk counts the colors */
    for(i = 0; i < Nbp; i++) {
      if(bpairs[i].col[1] == pk - 1) {           
	ii = bpairs[i].bi;
	ij = bpairs[i].bj;	  
	for(j = 0; j < i; j++) {
	  ji = bpairs[j].bi;
	  jj = bpairs[j].bj;
	  if (bpairs[j].out == -1) {
	    continue;
	  }
	 /*  if((bpairs[j].col[1] == pk -1)) {  */
	    if ((ii==ji)||(ii==jj)||(ij==ji)||(ij==jj)) {
	      bpairs[i].out = -1;
	      continue;
	  /*   } */
	    
	    if(((ji < ii) && (ii < jj) && (jj < ij)) ||
	       ((ii < ji) && (ji < ij) && (ij < jj))) {
	      bpairs[i].col[1] = pk;
	      mc = 1; /* found a new colour */
	    }
	  }
	}
      }     
    }
    
  }
  
 /*  printf("\nNO MWM:\n"); */
   for(i = 0; i < Nbp; i++) {
     if( (bpairs[i].out >= 0) &&
	 ( (bpairs[i].col[1] == 0) || (bpairs[i].col[1] == 1) ) ) {
       bpairs[i].out = 1;
     }
      /* if (bpairs[i].out > 0) {  */
/*        printf("%d-%d: wt=%d,col = %d, out=%d, comp=%d, mut=%d\n", */
/* 	      bpairs[i].bi, bpairs[i].bj, bpairs[i].wt, bpairs[i].col[1], */
/* 	      bpairs[i].out ,bpairs[i].comp , bpairs[i].mut ); */
/*       }  */
   }
   qsort(bpairs, Nbp, sizeof(INFO), cmp_bj);
   qsort(bpairs, Nbp, sizeof(INFO), cmp_bi);
}

/***************************************************************************/
INFO *choose_bp(INFO *bpairs) {

  int i, k, bct, bpc;
  double wwbpmax, wwbp;
  INFO *tpairs;

  
  wwbpmax = 0;
  wwbp = 0;
  bpc = 0;
  for(i = 0; i <  Nbp; i++){
    if (bpairs[i].wt >0) {
      bpc++;
      wwbp += bpairs[i].wt;
    }
    if(bpairs[i].wt > wwbpmax)
      wwbpmax = bpairs[i].wt;
  }  
  wwbp = wwbp/bpc;
  DW = 4*(int)(wwbp/100);
  MW = 4*(int)(wwbpmax/100);
 
    /* don't consider pairs with low weight */
  for(i=0; i<Nbp; i++) {    
    if ( bpairs[i].wt < (int)(Tresh*4*Pi_star)) /* 15.2. 14:45 100 statt wwbp */
      bpairs[i].wt = -2000;
     else if(bpairs[i].xx == 0)  
       bpairs[i].wt = -2000;  
  }

  for(i=0; i<Nbp; i++) {
    bpairs[i].wt = 4*(int)(bpairs[i].wt/4);
  }
  
  E = 0;
  for(i = 0; i <  Nbp; i++) {
    /* test if bpairs is o.k. */
    if((bpairs[i].bi == 0) || (bpairs[i].bj ==0))      
      fatal("choose_bp: in bpairs gibts base 0\n");
    /* count the relevant bps */
    if( (bpairs[i].wt > 0) && (bpairs[i].comp >= Min_comp)
       	&&(bpairs[i].ptyp[7] <= max_gap))
      E++;
  }
   
  /* "delete" the non-relevant bps */
  if(!(tpairs = (INFO *)calloc(E, sizeof(INFO)))) {
    fatal("choose_bp(): allocation of bpairs failed!!!\n");
  }
 
  bct = 0;
 
  for(i = 0; i <  Nbp; i++) {
    if( (bpairs[i].wt > 0) && (bpairs[i].comp >= Min_comp) 
	&&(bpairs[i].ptyp[7] <= max_gap)) {
      
      tpairs[bct].bi = bpairs[i].bi;
      tpairs[bct].bj = bpairs[i].bj;
      tpairs[bct].wt = bpairs[i].wt;
      tpairs[bct].owt = bpairs[i].owt;
      tpairs[bct].comp = bpairs[i].comp;
      tpairs[bct].mut = bpairs[i].mut;
      for(k=0; k<9; k++) 
	tpairs[bct].ptyp[k] = bpairs[i].ptyp[k];      
      bct++;
    }
  }
  
  free(bpairs);
  Nbp = bct;
  bpairs = tpairs;
  
  return bpairs;
}

/***************************************************************************/
void ini_bp(void) {
  int i, calsize;
  
  Nbp = 0;
  if(!(bpairs = (INFO *)calloc(P_BASE, sizeof(INFO)))) {
    fatal("helix(): allocation of bpairs failed!!!");
  }
  
  if(!(indx =  (int *) calloc((V+1), sizeof(int)))) {
    fatal("bp_together(): allocation of indx failed!!!");
  }
  for (i = 1; i <= V; i++) {
    indx[i] = (i*(i-1))/2;
  }
  calsize = (V * (V+1) ) / 2;
  if( !(Index = (int *) calloc(calsize, sizeof(int)) ) ) {
    fatal("bp_together(): allocation of Index failed!!!");
  }
}
/***************************************************************************/


/* End of file */
