/* Last changed Time-stamp: <2004-09-10 13:08:41 xtina> */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "xutils.h"
#define PRODEF
#include "hxmatch.h"
#include "read_graph.h"
#undef PRODEF

extern void read_graph(INFO *bpairs);
extern int  *Weighted_Match(void);
extern INFO *post_process(INFO *bpairs, int it);
static FILE *process_opts(int argc, char *argv[]);
static void use_me_this_way(void);


static int H_len;

int main (int argc, char **argv) {
  
  INFO *bpairs;
  INFO *alif_bp;
  FILE *file; 
  int i;
  
  /* set the parameters */
  phi1 = 0.8;
  phi2 = 60;
  AA = 75; 
  P_min = 0.90;
  Pi_star = 25;
  
  /* initialize options */
  H_len = 3; /* default lower-bond of helix-length */
  nomwm = 0; /* default: use MWM */ 
  Tresh = 1; /* default threshold */

  /*initialize variables */  
  af = 0;
  ap = 0;
  w = 0;
  Stop = 0;
  DW = 0;
  MW = 0;
  alif_bp = NULL;
  
  
   
  file =  process_opts(argc, argv);
  
  read_clustal(stdin); /*read the alignment and write the sequences
			 to global array Aln_seqs[] */
  
  bpairs = helix(H_len); /* generate the scoring matrix(=bpairs) */

  if(af) {
    alif_bp = alif(file);
    bpairs = mix_bpairs(alif_bp, bpairs);  
    free(alif_bp);
  }
  else if (ap) {
    alif_bp = alp(file);
    bpairs = mix_bpairs(alif_bp, bpairs);  
    free(alif_bp);
  }
  else if(w) {
    alif_bp = ext_pairs(file);
    bpairs = mix_bpairs(alif_bp, bpairs);  
    free(alif_bp);
  }

  bpairs = choose_bp(bpairs);
  /* print_roth(bpairs); */
  if (nomwm) {
    no_mwm(bpairs); 
    bpairs =  post_process(bpairs, 1);
  }     
  else { /* MWM */ 
    for(it = 1; it <= 30; it++) {
      read_graph(bpairs); /* fill the arrays needed by MWM */
      fprintf(stderr, "MWM: iteration %d\n", it);
      MATE = Weighted_Match(); /* MWM */
      fprintf(stderr, "postprocessing ...\n");
      bpairs =  post_process(bpairs, it); /* postprocessing and output */
      
      free(outcom);
      outcom = NULL;
      free(MATE);
      MATE = NULL;
      
      if(Stop) /* global variable - changed in post_process() */
	  break;
    }
  }

  for(i=1; i<=Nseq; i++) 
    free(Aln_seqs[i]);
  free(bpairs);
  
  return 0;
}


 /* -------------------------------------------------------- */
static FILE *process_opts(int argc, char *argv[]) {

  FILE *file = NULL;
  int i;
  
  for (i = 1; i < argc ; i++) {
    if (argv[i][0]=='-') {
      switch(argv[i][1]) {
      case 'l':
        if(strcmp(argv[i], "-len") == 0) {
	  sscanf(argv[++i], "%d", &H_len);
	  if (H_len <= 0) fatal("helix_len must be int >= 1!!!");
	}
        break;
      case 'T':
        if(strcmp(argv[i], "-T") == 0) {
	  sscanf(argv[++i], "%d", &Tresh);
	}	
	break;
      case 'P':
	if(strcmp(argv[i], "-PT") == 0) {
	  sscanf(argv[++i], "%d", &PTresh);
	}
	break;
      case 'N':
        if(strcmp(argv[i], "-N") == 0) {
	  nomwm = 1;
	}
	break;
      case 'A':
	if(strcmp(argv[i], "-A") == 0) {      
	  af = 1;
	  if(!(file = fopen (argv[++i], "r")))
	    fatal("no file alirna.ps\n");	  
	}
	else if(strcmp(argv[i], "-AP") == 0) {      
	  ap = 1;
	  if(!(file = fopen (argv[++i], "r")))
	    fatal("no file alifold.out\n");	  
	}
	break;     	
	case 'W':
	if(strcmp(argv[i], "-W") == 0) {
	  w = 1;
	  if(!(file = fopen (argv[++i], "r")))
	    fatal("no file!!!\n");	  
	}
	break;
      case 'h':
	use_me_this_way();
        break;
      default:
        use_me_this_way();
      }
    }
    else {
      
    }
  }
  return (file);
}

/* ------------------------------------------------------------- */

static void use_me_this_way(void) {
  fprintf(stderr, "\nhxmatch [options] < FOO > FOO.out\n");
  fprintf(stderr, "options:\n");
  fprintf(stderr, " -len <number>  set lower-bond of helix-length to <number>;\n (number has to be an integer)\n  (default: len 3)\n");
  fprintf(stderr, " -T <number> use threshold <number>*(pi*); (number has to be an integer, pi*=25)\n  (default T 1)\n");
  fprintf(stderr, " -N use the greedy algorithm instead of MWM\n  (default: MWM)\n");
  fprintf(stderr, " -A <alirna.ps> incorporate RNAalifold-results: all base pairs predicted by RNAalifold\n");
 fprintf(stderr, " -AP <alifold.out> incorporate RNAalifold-results: only base pairs with base pairing probability higher than 90 \%\n");
  fprintf(stderr, " -W <bpairs.list> incorporate weighted base pairs\n");
  exit (1);
}

/* -------------------------------------------------------------- */

/* End of file */










