/* Last changed Time-stamp: <2004-09-14 16:04:34 xtina> */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "xutils.h"
#include "hxmatch.h"
#include "read_graph.h"
/* #include "/usr/local/include/dmalloc.h" */

#define OUT_INC 10
#define MAL 0.9

typedef struct _col {
  int stlen;    /*if the color contains at least one stack with len > 2,
		  then stlen = 1, otherwise stlen = 0 */
  int cwt;      /* weight of the color */
  int cl[2];    /* cl[0]...old color, cl[1]...new color */
} COL;

typedef struct _stack {
  int bi, bj;
  int len, wt, col[2];
} STACK;

static void mate_to_outcom(INFO *bpairs);
static void scan_stacks(void);
static INFO *extend_stacks(INFO *bpairs);
static int cmp_fct2(const void *t1, const void *t2);
static int cmp_xx(const void *t1, const void *t2);
static void color(void);
static void print_cbrack(INFO *bpairs);
static void stacks(void);
static int cmp_stwt(const void *sw1, const void *sw2);
static int cmp_stbi(const void *si1, const void *si2);
static void color_outcom(void);
static INFO *bpair_new(INFO *bpairs);
static int cmp_bi(const void *i1, const void *i2);
static int cmp_bj(const void *j1, const void *j2);
static void nomwm_to_outcom(INFO *bpairs);

static int isbp;
static int OUT_top;
static int pk;
static STACK *stack;
static int ST; /* number of stacks in outcom */


INFO *post_process(INFO *bpairs, int it) {

  int k, i, j, bi, bj;

  if(nomwm){
    nomwm_to_outcom(bpairs);
    Stop =1;
  }
  else {    
  mate_to_outcom(bpairs);
  }
  
  scan_stacks(); /* mark isolated basepairs (xx = 0),
		    mark the beginnining(xx = 2) and
		    end of stacks (xx = 3) */
  outcom = extend_stacks(bpairs);

  /* reset .xx and scan again to get it right */
  for(i = 0; i < K; i++) {
    if(outcom[i].xx) {
      outcom[i].xx = 1;
    }
  }
  qsort(outcom, K, sizeof(INFO), cmp_xx);
  qsort(outcom, K, sizeof(INFO), cmp_fct2);
  scan_stacks();
  qsort(outcom, K, sizeof(INFO), cmp_fct2);

  /* write the information of consistent and inconsistent mutations
     from bpairs to outcom */ 
  for(i = 0; i < K; i++) {
    bi = outcom[i].bi;
    bj = outcom[i].bj;
    for(j = 0; j < Nbp; j++) {
      if((bi == bpairs[j].bi) && (bj == bpairs[j].bj)) {
	outcom[i].comp = bpairs[j].comp;
	outcom[i].mut = bpairs[j].mut;
	for(k=0; k<9; k++) {
	  outcom[i].ptyp[k] = bpairs[j].ptyp[k];
	}
	break;
      }
    }
  }

  /* set .xx = 0 if stacklen<3 */
  for(i = 0; i < K; i++) {
    if((outcom[i].xx ==2) && (outcom[i+1].xx == 3) ) {
      outcom[i].xx = 0;
      outcom[i+1].xx = 0;
      isbp += 2;
    }
  }
   
  /* set .xx = 0 if low weight */
  for(i = 0; i < K; i++) {
    if(outcom[i].wt < PTresh*25){
      outcom[i].xx =0;
    }
  }
  /* reset .xx and scan again to get it right */
  for(i = 0; i < K; i++) {
    if(outcom[i].xx) {
      outcom[i].xx = 1;
    }
  }
  qsort(outcom, K, sizeof(INFO), cmp_fct2);
  qsort(outcom, K, sizeof(INFO), cmp_xx);
  scan_stacks();
  qsort(outcom, K, sizeof(INFO), cmp_fct2);

  stacks(); /* combine the basepairs to stacks */
  qsort(stack, ST, sizeof(STACK), cmp_stwt);
  color(); /* find the color of the stacks */
  qsort(stack, ST, sizeof(STACK), cmp_stbi); 
  color_outcom(); /* assign the colors to the basepairs */
  
  print_cbrack(bpairs);
  
  bpairs = bpair_new(bpairs);
  
  free(stack);
  
  return bpairs;
}


/* -------------------------------------------------------*/

/* outcom contains basepairs (i,j) with i < j, and i1 < i2 < ... < iK
   first basepair in outcom[0] */

void mate_to_outcom(INFO *bpairs) {

  int  k;
  int i;

  K = 0;
  for(i = 1; i <= V; i++) {
    if(MATE[i] != 0)
      K++;
  }
  if(K%2)
    fatal("mate_to_outcom: K is not even!\n");
  K = K/2; /* MATE contains each pair twice, outcom only once */

  if(!(outcom = (INFO *)calloc(K, sizeof(INFO))))
    fatal("mate_to_outcom(): allocation of outcom failed!!!");

  k = 0;
  for(i = 1; i <= V; i++) {
    if(i < MATE[i]) {
      outcom[k].bi = i;
      outcom[k].bj = MATE[i];
      outcom[k].xx = 1;
      outcom[k].wt = out_weight[i];
      outcom[k].rm = RM[i];
      k++;
    }
   
  }
  free(out_weight);
  out_weight = NULL;
  free(RM);
  RM = NULL;
  
}

/* -----------------------------------------------------------*/
void scan_stacks(void) {
  
  /* skip isolated basepairs (xx = 0), mark the beginnining(xx = 2) and
     end of stacks (xx = 3) */
  
  int bst, est; /* to mark the beginning resp. end of a stack */
  int i;
  
  isbp = 0;
  /* are the first or the last basepair in the list isolated ? */
  if(!((outcom[1].bi == outcom[0].bi+1) &&
       (outcom[1].bj == outcom[0].bj-1))) {
    outcom[0].xx = 0;
    isbp++;
  }
  else outcom[0].xx = 2;  
  if(!((outcom[K-2].bi == outcom[K-1].bi-1) &&
       (outcom[K-2].bj == outcom[K-1].bj+1))) {
    outcom[K-1].xx = 0;
    isbp++;
  }
  else outcom[K-1].xx = 3;
  
  for(i = 1; i < K-1; i++) {
    if(!(outcom[i].xx)) continue;
    bst = est = 0;
    /* is the basepair the beginning of a stack? */
    if(!((outcom[i-1].bi == outcom[i].bi-1) &&
	 (outcom[i-1].bj == outcom[i].bj+1))) {
      outcom[i].xx = 2;
      bst =1;
    }
    /* is the basepair the end of a stack? */
    if( !((outcom[i+1].bi == outcom[i].bi+1) &&
	  (outcom[i+1].bj == outcom[i].bj-1))) {
      outcom[i].xx = 3;
      est = 1;
    }
    /* if the basepair is isolated, set xx = 0 */
    if(bst && est) {
      outcom[i].xx = 0;
      isbp++;
    }
  }
  
 qsort(outcom, K, sizeof(INFO), cmp_xx);
   
}


/* -------------------------------------------------------*/

static INFO *extend_stacks(INFO *bpairs)  {

  /* now extend stacks, if possible */  
  int bi, bj, extend;
  int i, j, k;

  OUT_top = K;
  for(i = 0; i < K; i++) {
    switch (outcom[i].xx) {
    case 0:
      break;
    case 1:
      break;
    case 2:
      bi = outcom[i].bi - 1;
      bj = outcom[i].bj + 1;
      extend = 1;
      /* if the basepair extending the stack is contained in the
	 list of all possible pairs ... */
      for(k = 0; k < Nbp; k++) {  
	if((bi == bpairs[k].bi) && (bj == bpairs[k].bj) &&
	   (bpairs[k].wt > 0) && (bpairs[k].comp >= Min_comp)) {
	  /* ...have a look whether one of the bases is already
	     contained in outcom ... */
	  for(j = 0; j < K; j++) {
	    if((outcom[j].xx) &&
	       ((bi == outcom[j].bi) || (bi == outcom[j].bj) ||
		(bj == outcom[j].bi) || (bj == outcom[j].bj) )) {
	      /* ...if outcom already contains one of the bases,
		 don't take the bp */
	      extend = 0;
	      break;
	    }
	  }
	  /* put the bp into outcom */
	  if(extend) {
	    if (K >= OUT_top) {
	      if (!(outcom = (INFO *)realloc(outcom,
					     (OUT_top+OUT_INC)*sizeof(INFO))))
		fatal("stack_verl(): realloc of outcom failed!!!");
	      cset_info(outcom, K, K +OUT_INC);	     
	      OUT_top += OUT_INC;
	    }
	    outcom[K].bi = bi;
	    outcom[K].bj = bj;
	    outcom[K].xx = outcom[i].xx;
	    outcom[K].col[0] = 0;
	    outcom[K].col[1] = 0;
	    outcom[K++].wt = bpairs[k].wt;
	    /*the former opening bp is now inside a stack, therefore xx = 1 */
	    outcom[i].xx = 1;	   	      
	  }
	}
      }
      break;
      
    case 3:
      bi = outcom[i].bi + 1;
      bj = outcom[i].bj - 1;
      extend = 1;
      /* if the basepair extending the stack is contained in the
	 list of all possible pairs ... */
      for(k = 0; k < Nbp; k++) {  
	if((bi == bpairs[k].bi) && (bj == bpairs[k].bj) &&
	   (bpairs[k].wt > 0) && (bpairs[k].comp >= Min_comp)) {
	  /* ...have a look whether one of the bases is already
	     contained in outcom ... */
	  for(j = 0; j < K; j++) {
	    if((outcom[j].xx) &&
	       ((bi == outcom[j].bi) || (bi == outcom[j].bj) ||
		(bj == outcom[j].bi) || (bj == outcom[j].bj) )) {
	      /* ...don't take the bp */
	      extend = 0;
	      break;
	    }
	  }
	  /* put the bp into outcom */
	  if(extend) {
	    if (K >= OUT_top) {
	      if (!(outcom = (INFO *)realloc(outcom,
					     (OUT_top+OUT_INC)*sizeof(INFO))))
		fatal("stack_verl(): realloc of outcom failed!!!");	   
	      cset_info(outcom, K, K+OUT_INC);
	      OUT_top += OUT_INC;
	    }
	    outcom[K].bi = bi;
	    outcom[K].bj = bj;
	    outcom[K].xx = outcom[i].xx;
	    outcom[K].col[0] = 0;
	    outcom[K].col[1] = 0;
	    outcom[K++].wt = bpairs[k].wt;
	    /*the former closing bp is now inside a stack, therefore xx = 1 */
	    outcom[i].xx = 1;
	  }
	}
      }
      break;

    default:
      fprintf(stderr, "warning: unknown tag %d in s_elongation\n",
	      outcom[i].xx);
    }
  }
  
return outcom;
}

/* -------------------------------------------------------*/

static int cmp_xx(const void *t1, const void *t2) {
  return ((INFO *)t1)->xx - ((INFO *)t2)->xx;
}

/* -------------------------------------------------------*/

static int cmp_fct2(const void *t1, const void *t2) {
   if(!((INFO *)t1)->xx)
     return 0;

   return  ((INFO *)t1)->bi - ((INFO *)t2)->bi;
 }

/* -------------------------------------------------------*/

static int cmp_stwt(const void *sw1, const void *sw2) {
  return ((STACK *)sw2)->wt - ((STACK *)sw1)->wt;
}

/* -------------------------------------------------------*/

static int cmp_stbi(const void *si1, const void *si2) {
  return ((STACK *)si1)->bi - ((STACK *)si2)->bi;
}

/* -------------------------------------------------------*/

void color(void) {

  int i, j;
  int ii, ij, ji, jj;
  int mc;


  /* color the bp's (different colors are 0, 1, 2,... in outcom[i].col[0]) */
  pk = 0;
  mc = 1;
  while(mc) { /* if mc = 0 after a search, there are no more colors*/
    mc = 0;
    pk++; /* pk counts the colors */
    for(i = 0; i < ST; i++) {
      if(stack[i].col[0] == pk - 1) {           
	ii = stack[i].bi;
	ij = stack[i].bj;
	for(j = 0; j < i; j++) {
	  if((stack[j].col[0]== pk -1)) { 
	    ji = stack[j].bi;
	    jj = stack[j].bj;
	    if(((ji < ii) && (ii < jj) && (jj < ij)) ||
	       ((ii < ji) && (ji < ij) && (ij < jj))) {
	      stack[i].col[0] = pk;
	      mc = 1; /* found a new colour */
	    }
	  }
	}
      }
    }
  }
 
  
}


/* -------------------------------------------------------*/

void stacks(void) {

  int i, k;
  int len, wt;

  ST = 0;
  for(i = isbp; i < K; i++) {
    if(outcom[i].xx == 2)
      ST++;
  }

  if(!(stack = (STACK *)calloc(ST, sizeof(STACK))))
    fatal("main(): allocation of stack failed!!!");

  k = 0;
  for(i = isbp; i < K; i++) {
    len = 0;
    wt = 0;
    if(outcom[i].xx == 2) {
      outcom[i].hx = k;
      stack[k].bi = outcom[i].bi;
      stack[k].bj = outcom[i].bj;
      len++;
      wt += outcom[i].wt;
      while(outcom[i+len].xx == 1) {
	outcom[i+len].hx = k;
	wt += outcom[i+len].wt;
	len++;
      }
      outcom[i+len].hx = k;
      wt += outcom[i+len].wt;
      len++;
      stack[k].len = len;
      stack[k].wt = wt;
      k++;
    }
  }
}

/* -------------------------------------------------------*/

void color_outcom(void) {

  int i, k;
  
  for(k=0; k<ST; k++) {
    for(i=isbp; i<K; i++) {
      if(stack[k].bi == outcom[i].bi) {
	outcom[i].col[0] = stack[k].col[0];
	i++;
	while(outcom[i].xx != 2) {
	  outcom[i].col[0] = stack[k].col[0];
	  i++;
	  if(i == K) break;
	}
      }
    }
  }  
}
/* -------------------------------------------------------*/

void print_cbrack(INFO *bpairs) {

  int i;
  int ii, ij;
  char *b;
  char *oldb;
  char col;
  FILE *file;
  int verschieden;

  
  oldb = Struc;
 
  
  if(!(b = (char *) calloc(V, sizeof(char))))
    fatal("print_cbrack: allocation of b failed!\n");
  
  
  for(i = 0; i < V; i++)
    b[i] = '.';

  for(i = 0; i < K; i++) {
    if(outcom[i].xx) {
      ii = outcom[i].bi-1;
      ij = outcom[i].bj-1;
      switch(outcom[i].col[0]) {
      case 0:
	b[ii] = '(';
	b[ij] = ')';
	break;
      case 1:      
	b[ii] = '[';
	b[ij] = ']';
	break;
      case 2:
	b[ii] = '{';
	b[ij] = '}';
	break;
      case 3:
	b[ii] = '<';
	b[ij] = '>';
	break;
      case 4:
	b[ii] = 'A';
	b[ij] = 'a';
	break;
      case 5:
	b[ii] = 'B';
	b[ij] = 'b';
	break;
      case 6:
	b[ii] = 'C';
	b[ij] = 'c';
	break;
      case 7:
	b[ii] = 'D';
	b[ij] = 'd';
	break;
      case 8:
	b[ii] = 'E';
	b[ij] = 'e';
	break;
      case 9:
	b[ii] = 'F';
	b[ij] = 'f';
	break;
      case 10:
	b[ii] = 'G';
	b[ij] = 'g';
	break;
      default:
	b[ii] = '!';
	b[ij] = '!';
	break;
      }
    }

  } 

 
 
    if(oldb) {
      verschieden = 0;
      for(i = 0; i < V; i++) {
	if (b[i] != oldb[i]) 
	  verschieden =1;
      }
      if(!verschieden) {
	Stop =1;
      }
    }
  
    if(Stop) { 
      
      printf("\n  i    j     wt     xx  col    RM  comp g-g mut\n"); 
    for(i = 0; i < K; i++) {
      if(outcom[i].col[0])
	col = 'L';
      else
	col = 'U';
      if(outcom[i].xx) 
	printf("%4d %4d  %5d   %2d    %c   %3d  %3d   %2d   %d\n",
	       outcom[i].bi, outcom[i].bj, (outcom[i].wt)/4, outcom[i].xx,
	       col, outcom[i].rm, outcom[i].comp,
	       outcom[i].ptyp[7], outcom[i].mut );    
    }
#if 0    
    for(i = 0; i < V; i++) 
      printf("%c", b[i]);    
    printf("\n");
#endif    
    /* print the "mfe"-file */    
    file = fopen("name_db.hx", "w");
    fprintf(file, "name\n");
    for(i = 0; i < V; i++)
      fprintf(file, "%c", Cons_seq[i]);
    fprintf(file, "\n");      
    for(i = 0; i < V; i++)
      fprintf(file, "%c", b[i]);
    fprintf(file, "\n");
    fclose(file);
   }
    
  Struc = b;

  free(oldb);
  oldb = NULL;
  if(Stop){
    free(b);
    b = NULL;
    Struc = NULL;
    free(Cons_seq);
  }
  
} 
/***********************************************************************/

static INFO *bpair_new(INFO *bpairs) {
  INFO *tpairs; /* temporary bpairs */
  int i, j, rbp, ct;
  int ii, ij, ji, jj;
  int conflict0, conflict1;

  for(i = isbp; i < K; i++) {
    if(outcom[i].col[0] > 1)
      outcom[i].xx = 0;
  }
  for(i=0; i<Nbp; i++) {
    bpairs[i].xx = 1;
  }

  /* chuck out bpairs that have color > 1 corresponding to outcom */
   for(i=0; i<Nbp; i++) {
     if(bpairs[i].xx) {
       ii = bpairs[i].bi;
       ij = bpairs[i].bj;
       conflict0 = 0;
       conflict1 = 0;
       for(j=0; j<K; j++) {	
	 if(!(outcom[j].xx)) continue;
	 ji = outcom[j].bi;
	 jj = outcom[j].bj;
	 if(((ji < ii) && (ii < jj) && (jj < ij)) ||
	    ((ii < ji) && (ji < ij) && (ij < jj))) {
	   if(outcom[j].col[0] == 0)
	     conflict0 = 1;
	   if(outcom[j].col[0] == 1)
	     conflict1 = 1;
	 }	 
       }
       if((conflict0) && (conflict1))
	 bpairs[i].xx = 0;
       else if (conflict0)
	 bpairs[i].col[0] = 1;
     }
   }

   
  /* count remaining bp's */
  rbp = 0;
  for(i=0; i<Nbp; i++) {
    if(bpairs[i].xx)
      rbp++;
  }
  if(!(tpairs = (INFO *)calloc(rbp+1, sizeof(INFO)))) {
    fatal("bpair_new(): allocation of tpairs failed!!!");
  }
  ct = 0;
  for(i=0; i<Nbp; i++) {
    if(bpairs[i].xx) {
      tpairs[ct].bi = bpairs[i].bi ;
      tpairs[ct].bj = bpairs[i].bj ;
      tpairs[ct].wt = bpairs[i].wt ;
      tpairs[ct].owt = bpairs[i].owt ;
      tpairs[ct].comp = bpairs[i].comp ;
      tpairs[ct].mut = bpairs[i].mut ;
      tpairs[ct].col[0] = bpairs[i].col[0];
      for(j=0; j<9; j++) {
	tpairs[ct].ptyp[j] = bpairs[i].ptyp[j] ;
      }
      tpairs[ct].xx = -1 ;
      ct++;
    }
  }
  
  free(bpairs);
  bpairs = tpairs;
  Nbp = rbp;

  qsort(bpairs, Nbp, sizeof(INFO), cmp_bj);
  qsort(bpairs, Nbp, sizeof(INFO), cmp_bi);
  
  
  return bpairs;
}
/***********************************************************************/
INFO *cset_info(INFO *inf, int f, int p) {
  /* reset the memory for INFO *inf to zero, from element f until element p-1 */ 
  int i, l;
  
  for(i=f; i<p; i++) {
    inf[i].bi = 0;
    inf[i].bj = 0;
    inf[i].wt = 0;
    inf[i].owt = 0;
    inf[i].xx = 0;
    inf[i].rm = 0;
    inf[i].comp = 0;
    inf[i].mut = 0;
    inf[i].hx = 0;
    inf[i].out = 0;
    inf[i].col[0] = 0;
    inf[i].col[1] = 0;
    for(l=0; l<9; l++) {
      inf[i].ptyp[l] =0;
    }
  }
  return inf;
}
/***********************************************************************/
static int cmp_bi(const void *i1, const void *i2) {
  return ((INFO *)i1)->bi - ((INFO *)i2)->bi;
}
static int cmp_bj(const void *j1, const void *j2) {
  return ((INFO *)j1)->bj - ((INFO *)j2)->bj;
}
/***********************************************************************/
/* -----------------------------------------------------------*/
void nomwm_to_outcom(INFO *bpairs) {
  int  k;
  int i;

  K = 0;
  for(i = 0; i < Nbp; i++) {
    if (bpairs[i].out > 0) {
      K++;
    }
  }
  if(!(outcom = (INFO *)calloc(K, sizeof(INFO))))
    fatal("mate_to_outcom(): allocation of outcom failed!!!");
  
  k = 0;

  for(i = 0; i < Nbp; i++) {
    if (bpairs[i].out > 0) {
      outcom[k].bi = bpairs[i].bi;
      outcom[k].bj = bpairs[i].bj;
      outcom[k].xx = 1;
      outcom[k].wt = bpairs[i].wt;
      outcom[k].rm = 0;
      outcom[k].col[1] = bpairs[i].col[1];
      k++;
    }
  }
  
}

/* End of file */

