/* Last changed Time-stamp: <2004-09-10 11:20:42 xtina> */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "xutils.h"
#include "hxmatch.h"
#include "read_graph.h"

static int cmp_fct(const void *t1, const void *t2);

void read_graph(INFO *bpairs) {

  int i, j, wt, ct, ce, allocsize, tt;
  
  E = 0;
  for(i = 0; i <  Nbp; i++) {
    if( (bpairs[i].wt > 0) && (bpairs[i].comp >= Min_comp))
      E++;
  }
  
  for(i = 0; i < Nbp; i++) {
    if(bpairs[i].bi > bpairs[i].bj) {
      tt = bpairs[i].bi;
      bpairs[i].bi =  bpairs[i].bj;
      bpairs[i].bj = tt;
    }
    if((bpairs[i].bi == 0) ||(bpairs[i].bj == 0))
      fatal("read_graph: there is a vertex '0'!!!\n");
  }
  
  qsort(bpairs, Nbp, sizeof(INFO), cmp_fct);
  
  allocsize = V + 2*E + 3;
  
  A = calloc (allocsize, sizeof(int));
  END = calloc (allocsize, sizeof(int));
  WEIGHT = calloc (allocsize, sizeof(int));
  RM = calloc(V+1, sizeof(int));
  
  if(!(A && END && WEIGHT && RM))
    fatal("read_graph(): allocation of A, END or WEIGHT or RM failed!!!");
  
  ce = V+2;
  
  for (ct = Nbp-1; ct >= 0; ct--){
    wt=bpairs[ct].wt;
    if( (wt > 0) && (bpairs[ct].comp >= Min_comp)) {
      j=bpairs[ct].bi;
      i=bpairs[ct].bj;
      WEIGHT[ce-1] = WEIGHT[ce] = wt;
      END[ce-1] = i;
      END[ce] = j;
      A[ce] = A[i];
      A[i] = ce;
      A[ce-1] = A[j];
      A[j] = ce-1;
      ce += 2;
    }
  }
}
/********************************************************/
static int cmp_fct(const void *t1, const void *t2) {

  INFO *tp1, *tp2;
  tp1 = (INFO *)t1;
  tp2 = (INFO *)t2;
  
  if(tp1->bi != tp2->bi) return tp1->bi - tp2->bi;
  return tp1->bj - tp2->bj;
}
/********************************************************/

void read_clustal(FILE *clust) {
  char *line, workstr[100], *strg;
  int i, j, n;
  char b[] = "NACGU-";
  int conschar[6];
  char cc;
  int max_c, k, nq;

  fprintf(stderr, "reading clustal file\n");
  line=getline(clust);
  sscanf(line,"%99s",workstr);
  
  if(strcmp(workstr,"CLUSTAL") !=0) {
    fatal("read_clustal: Sorry. Expecting CLUSTAL file.\n");
  }

  free(line);
  line = getline(clust);
  while ( (line!=NULL) && ((line[0]=='\0') || isspace(line[0])) ){
    free(line);
    line = getline(clust); 
  }  
  
  Nseq=0;
  while (!isspace(line[0]) ) {
    n = strlen(line);
    if(! (strg = (char *) calloc( (n+1), sizeof(char) )))
      fatal("read_clustal(): allocation of strg failed!!!");
    sscanf(line, "%s %s", workstr, strg);
    Nseq++;
    
    if(! (Names[Nseq] = (char *) calloc(strlen(workstr)+1, sizeof(char)) ) )
      fatal("read_clustal(): allocation of Names[Nseq] failed!!!");
    strcpy(Names[Nseq],workstr); 
    
    if(! (Aln_seqs[Nseq] =  (char *)calloc(strlen(strg)+1,  sizeof(char) ) ))
      fatal("read_clustal(): allocation of Aln_eqs[Nseq]  failed!!!");
    strcat(Aln_seqs[Nseq],strg);
    free(strg);
    free(line);
    line = getline(clust);
  }
  
  while(line!=NULL) {
    
    while ( (line!=NULL) && ((line[0]=='\0') || isspace(line[0])) ){
      free(line);
      line = getline(clust); 
    }     
    
    nq=0;
    while ( (line!=NULL) && !(isspace(line[0])) ) {
      n = strlen(line);
      strg = (char *) calloc( (n+1), sizeof(char) );
      sscanf(line,"%s %s",workstr,strg);
      nq++;
      if ( strcmp(workstr, Names[nq]) == 0 ) {
	Aln_seqs[nq] = (char *)
	  realloc(Aln_seqs[nq], 
		  strlen(strg)+strlen(Aln_seqs[nq])+1);
	strcat(Aln_seqs[nq],strg);   
      } else {
	printf("nq=%d ws=%s\n n=%s\n", nq, workstr, Names[nq]);
	fatal ("Sorry, something wrong with your file.\n");
      }
      free(strg);
      free(line);
      line = getline(clust);
    }   
  }   
  free(line);

  n = strlen(Aln_seqs[1]); 
  for (i=2; i<=Nseq; i++) {
    if (strlen(Aln_seqs[i])!=n) {
      fatal("Sorry, something wrong with your file.\n"
	      "Unequal lengths!\n\n");
    }
  }
  
  for (i=1; i<=Nseq; i++) 
    for (j=0; j<n; j++)
      if (Aln_seqs[i][j]=='T') Aln_seqs[i][j]='U';

  fprintf(stderr, "%d sequences; length of alignment %d\n", Nseq, n);

  /* find consensus sequence */
  if(! (Cons_seq =  (char *)calloc(n+1,  sizeof(char) ) ))
    fatal("read_clustal(): allocation of Cons_seq failed!!!\n");
  for (j=0; j<n; j++) {
    for(k=0; k<=5;k++)
      conschar[k] = 0;
    
    for (i=1; i<=Nseq; i++) {
      switch (Aln_seqs[i][j]) {
      case 'N': 
	conschar[0]++;
	break;
      case 'A':
	conschar[1]++;
	break;
      case 'C':
	conschar[2]++;
	break;
      case 'G':
	conschar[3]++;
	break;
      case 'U':
	conschar[4]++;
	break;
      case '-':
	conschar[5]++;
	break;
      } 
    }
    cc = 'N';
    max_c = conschar[0];
    for(k=0; k<=5;k++) {
      if(conschar[k] > max_c) {
	max_c = conschar[k];
	cc = b[k];
      }
    }
    Cons_seq[j] = cc;
  }
  
  for (i=1; i<=Nseq; i++) 
    free(Names[i]);
  
  return;
}

/*-------------------------------------------------------------------------*/

/* End of file */
