/* Last changed Time-stamp: <2004-01-12 15:38:47 xtina> */

/* N-cubed weighted matching */
/* Implementation of H. Gabow's Ph.D. thesis, Stanford Univ. 1973 */
/* Written by Edward Rothberg  7/85 */
/* For complete details, please refer to the original paper */

/* updates numerical bounds for linking paths. */
/* called with LAST_D set to the bound on DELTA for the next search */

#include <stdio.h>
#include <stdlib.h>
#include "xutils.h"
#include "wmatch.h"
#include "read_graph.h"

void SET_BOUNDS(void) {
 
  int i, del;
  
  for (v=1; v <= V; ++v) {
    if (LINK[v] < 0 || BASE[v] != v) {
      NEXT_D[v] = LAST_D;
      continue;
    }
    LINK[v] = -LINK[v];
    i = v;
    while (i != DUMMYVERTEX) {
      Y[i] -= DELTA;
      i = NEXTVTX[i];
    }
    f = MATE[v];
    if (f != DUMMYEDGE) {
      i = BEND(f);
      del = SLACK(f);
      while (i != DUMMYVERTEX) {
	Y[i] -= del;
	i = NEXTVTX[i];
      }
    }
    NEXT_D[v] = LAST_D;
  }
}


/* undoes all blossoms to get the final matching */

void UNPAIR_ALL(void) {

  int u;

  for (v=1; v <= V; ++v) {
    if (BASE[v] != v || LASTVTX[v] == v)
      continue;
    nextu = v;
    NEXTVTX[LASTVTX[nextu]] = DUMMYVERTEX;
    while (1) {
      u = nextu;
      nextu = NEXTVTX[nextu];
      UNLINK (u);
      if (LASTVTX[u] != u) {
	f = (LASTEDGE[2] == OPPEDGE(e)) ? LASTEDGE[1] : LASTEDGE[2];
	NEXTVTX[LASTVTX[BEND(f)]] = u;
      }
      newbase = BMATE (BMATE(u));
      if (newbase != DUMMYVERTEX && newbase != u) {
	LINK[u] = -DUMMYEDGE;
	REMATCH (newbase, MATE[u]);
      }
      while (LASTVTX[nextu] == nextu && nextu != DUMMYVERTEX)
	nextu = NEXTVTX[nextu];
      if (LASTVTX[nextu] == nextu && nextu == DUMMYVERTEX)
	break;
    }
  }
}


