/* Last changed Time-stamp: <2004-01-12 15:40:08 xtina> */

/* N-cubed weighted matching */
/* Implementation of H. Gabow's Ph.D. thesis, Stanford Univ. 1973 */
/* Written by Edward Rothberg  7/85 */
/* For complete details, please refer to the original paper */
/* Expands a blossom.  Fixes up LINK and MATE. */

#include <stdio.h>
#include <stdlib.h>
#include "xutils.h"
#include "wmatch.h"
#include "read_graph.h"


void UNPAIR(int oldbase, int oldmate) {

  int  uu, ee;

#ifdef DEBUG
  printf("Unpair oldbase, oldmate=%d %d\n",oldbase, oldmate);
#endif
  
  UNLINK (oldbase);
  newbase = BMATE (oldmate);
  if (newbase != oldbase) {
    LINK [oldbase] = -DUMMYEDGE;
    REMATCH (newbase, MATE[oldbase]);
    if (f == LASTEDGE[1])
      LINK[secondmate] = -LASTEDGE[2];
    else LINK[secondmate] = -LASTEDGE[1];
  }
  ee = LINK[oldmate];
  uu = BEND (OPPEDGE (ee));
  if (uu == newbase) {
    POINTER (newbase, oldmate, ee);
    return;
  }
  LINK[BMATE (uu)] = -ee;

  do {
    ee = -LINK[uu];
    if((oldbase == 4166) && (oldmate == 5290)) {
      printf("1: uu=%d\n", uu);
      printf("1: ee=%d\n", ee);
    }
    v = BMATE (uu);
    if((oldbase == 4166) && (oldmate == 5290)) {
      printf("2: uu=%d\n", uu);
      printf("2: ee=%d\n", ee);
    }
    POINTER (uu, v, -LINK[v]);
    if((oldbase == 4166) && (oldmate == 5290)) {
      printf("3: uu=%d\n", uu);
      printf("3: ee=%d\n", ee);
    }
    uu = BEND (ee);
    if((oldbase == 4166) && (oldmate == 5290)) {
      printf("4: uu=%d\n", uu);
      printf("4: ee=%d\n", ee);
    }
  } while (uu != newbase);
  ee = OPPEDGE (ee);
  POINTER (newbase, oldmate, ee);
}


/* changes the matching along an alternating path */
/* firstmate is the first base vertex on the path */
/* edge e is the new matched edge for firstmate   */

void REMATCH(int firstmate, int e) {

#ifdef DEBUG
  printf("Rematch firstmate=%d e=%d-%d\n",firstmate, END[OPPEDGE(e)], END[e]);
#endif
  
  MATE[firstmate] = e;
  nexte = -LINK[firstmate];
  if (nexte != DUMMYEDGE) {
    RM[firstmate]++;
    RM[END[e]]++;
  }
  while (nexte != DUMMYEDGE) {
    e = nexte;
    f = OPPEDGE (e);
    firstmate = BEND (e);
    secondmate = BEND (f);
    RM[firstmate]++;
    RM[secondmate]++;
    nexte = -LINK[firstmate];
    LINK[firstmate] = -MATE[secondmate];
    LINK[secondmate] = -MATE[firstmate];
    MATE[firstmate] = f;
    MATE[secondmate] = e;
  }
}


/* unlinks subblossoms in a blossom.  oldbase is the base of the blossom to */
/* be unlinked. */

void UNLINK(int oldbase) {
  
  int k, i, j=1;
  
#ifdef DEBUG
  printf("Unlink oldbase=%d\n",oldbase);
#endif
  
  i = NEXTVTX[oldbase];
  newbase = NEXTVTX[oldbase];
  nextbase = NEXTVTX[LASTVTX[newbase]];
  e = LINK[nextbase];
 UL2:
  do {
    nextedge = OPPEDGE (LINK[newbase]);
    for (k=1; k <= 2; ++k) {
      LINK[newbase] = -LINK[newbase];
      BASE[i] = newbase;
      i = NEXTVTX[i];
      while (i != nextbase) {
	BASE[i] = newbase;
	i = NEXTVTX[i];
      } 
      newbase = nextbase;
      nextbase = NEXTVTX[LASTVTX[newbase]];
    }
  } while (LINK[nextbase] == nextedge);
  if (j==1) {
    LASTEDGE[1] = nextedge;
    j++;
    nextedge = OPPEDGE (e);
    if (LINK[nextbase] == nextedge) {
      goto UL2;
    }
  }
  LASTEDGE[2] = nextedge;
  
  if (BASE[LASTVTX[oldbase]] == oldbase) 
    NEXTVTX[oldbase] = newbase;
  else {
    NEXTVTX[oldbase] = DUMMYVERTEX;
    LASTVTX[oldbase] = oldbase;
  }
}


