/* Last changed Time-stamp: <2004-01-16 13:51:16 xtina> */

/* N-cubed weighted matching */
/* Implementation of H. Gabow's Ph.D. thesis, Stanford Univ. 1973 */
/* Written by Edward Rothberg  7/85 */
/* For complete details, please refer to the original paper */



#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "xutils.h"
#include "read_graph.h"
#include "hxmatch.h"

#define PRODEF
#include "wmatch.h"
#undef PRODEF

static void Initialize(void);
static void FreeUp(void);

int *Weighted_Match(void) {

  int g, i, j, w, outcome;
  /* int loop = 1; */

  /* set up internal data structure */
  Initialize();
  /*begin XTINA */
  fprintf(stderr, "V= %d, E = %d\n", V, E);
  /*end XTINA */
  for(;;) {
    DELTA = 0;
    for(v = 1; v <= V; ++v)
      if(MATE[v] == DUMMYEDGE)
	POINTER (DUMMYVERTEX, v, DUMMYEDGE);
    for(;;) {
      i = 1;
      for(j = 2; j <= V; ++j)
	if (NEXT_D[i] > NEXT_D[j])
	  i = j;
      DELTA = NEXT_D[i];
      if(DELTA == LAST_D)
	goto done;
      v = BASE[i];
      if (LINK[v] >= 0) {
	outcome = PAIR();
	if (outcome == 1)
	  break;
      }
      else {
	w = BMATE (v);
	if (LINK[w] < 0) {
	  POINTER (v, w, OPPEDGE(NEXTEDGE[i]));
 	}
	else UNPAIR (v, w);
      }
    }

    LAST_D -=DELTA;
    SET_BOUNDS();
    g = OPPEDGE(e);
    REMATCH (BEND(e), g);
    REMATCH (BEND(g), e);
#ifdef DEBUG
    printf("i\tA\tEND\tWEIGHT\tNEXTPAIR\n");
    for(i=1;i<=V+2*E+2; i++)
      printf("%d\t %d\t%d\t%d\t%d\n",i,A[i], END[i], WEIGHT[i], NEXTPAIR[i]);
    printf("i  MATE  LINK  BASE  NXTVTX  LSTVTX     Y    NEXT_D   NEXTEDGE REMATCHFLAGS\n");
    for(i=1;i<=V+1; i++)
      printf("%2d   %3d  %2d %2d %2d %2d %5d  %5d %3d\n",
	     i, MATE[i], LINK[i], BASE[i], NEXTVTX[i],
	     LASTVTX[i], Y[i],NEXT_D[i],NEXTEDGE[i] );
#endif
  }
  
 done:
  SET_BOUNDS();
  UNPAIR_ALL();
  for (i=1; i<=V;++i) {
    out_weight[i] = WEIGHT[MATE[i]];
    MATE[i] = END[MATE[i]];
    if (MATE[i]==DUMMYVERTEX)
      MATE[i]=0;
  }

  FreeUp();

#ifdef DEBUG
  printf("MAXIMUM WEIGHTED MATCHING");
  for(i=0; i<=V; i++)
    printf("%2d, %2d\n", i, MATE[i]);
#endif
  
  return(MATE);
}


void Initialize(void) {

  int i, allocsize;
  int max_wt = 0;

  DUMMYVERTEX =V +1;
  DUMMYEDGE = V+2*E+1;
  END[DUMMYEDGE] = DUMMYVERTEX;

  for (i=V+2; i<=V+2*E; i+=2) {
    if (WEIGHT[i] > max_wt)
      max_wt = WEIGHT[i];
  }

    LAST_D = max_wt/2;

    allocsize = (V+2)*sizeof(int);
    MATE     = (int *) malloc(allocsize);
    LINK     = (int *) malloc(allocsize);
    BASE     = (int *) malloc(allocsize);
    NEXTVTX  = (int *) malloc(allocsize);
    LASTVTX  = (int *) malloc(allocsize);
    Y        = (int *) malloc(allocsize);
    NEXT_D   = (int *) malloc(allocsize);
    NEXTEDGE = (int *) malloc(allocsize);
    allocsize = (V+2*E+3)*sizeof(int);
    NEXTPAIR = (int *) malloc(allocsize);
    out_weight = (int *) malloc(allocsize);
    
    for (i = 1; i <= V+1; i++) {
	MATE[i] = DUMMYEDGE;
	NEXTEDGE[i] = DUMMYEDGE;
	NEXTVTX[i] = 0;
	LINK[i] = -DUMMYEDGE;
	BASE[i] = i;
	LASTVTX[i] = i;
	Y[i] = LAST_D;
	NEXT_D[i] = LAST_D;
	out_weight[i] = 0;
	}
}

void FreeUp(void) {

    free(LINK);
    free(BASE);
    free(NEXTVTX);
    free(LASTVTX);
    free(Y);
    free(NEXT_D);
    free(NEXTEDGE);
    free(NEXTPAIR);
    free(A);
    free(END);
    free(WEIGHT);
}

